unit CodeEditor;

(***** BEGIN LICENSE BLOCK *****
Version: MPL 1.1

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is CodeEditor.pas

The Initial Developer of the Original Code is
Florian Kberle. (eMail:FloriansKarten@web.de)
Portions created by the Initial Developer are Copyright (C) 2005
the Initial Developer. All Rights Reserved.

Contributor(s):

***** END LICENSE BLOCK *****)

interface
uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  WortErkennungsUnit,stdctrls,CodeEditor_Hint,CodeEditorConst,Menus;
{$DEFINE ERRORCHECK}
type
  TIndexSpeicher=0..HighColorIndex;

  TOnFunctionHint=procedure(Funktion:String;var Hint:THintStrings) of object;
  //TCodeFont = class;
  TAktionsListe=class;
  TUndoList=class;
  TRepeatList=class;
  TCodeEditorConfig=class;

  TCodeEditor = class(TCustomControl)
  private //Link: TCodeEditor-Deklaration
    FCurX:Integer;
    FCurY:Integer;
    FSelX:Integer;
    FSelY:Integer;
    FUpdatingStrings:Boolean;
    FConfig:TCodeEditorConfig;
    FStrings:TStrings;
    //FCodeFont:TCodeFont;
    FFirstLine:Integer;
    FLastLine:Integer;
    FFirstChar:Integer;
    FLastChar:Integer;
    FTimerID:Integer;
    FVerschiebungsOrt:TPoint;
    FVerschiebungstattMarkierung:Boolean;
    FAbstandzumRand:Integer;
    FBlinkZeit:LongWord;
    FTextCursorVisible:Boolean;
    FDoubleClick:Boolean;
    FHintForm:THintForm;
    FHintCurPos:TPoint;
    FOnFunctionHint:TOnFunctionHint;
    FUndoList: TUndoList;
    FRepeatList: TRepeatList;
    FMouseDown:Boolean;
    procedure WMKeyDown(var Message:TWMKey);message WM_KEYDOWN;
    procedure WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
    procedure WMCreate(var Message: TWMCreate); message WM_CREATE;
    procedure WMVScroll(var Message: TWMVScroll); message WM_VSCROLL;
    procedure WMHScroll(var Message: TWMHScroll);  message WM_HSCROLL;
    procedure WMTimer(var Message :TWMTimer);message WM_TIMER;
    procedure WMDblClick(var Message:TWMLButtonDblClk );message WM_LButtonDblClk;
    procedure WMMoving(var Message:TMessage);message WM_MOVING;
    procedure WMCopy(var Message:TWMCopy);message WM_COPY;
    procedure WMPaste(var Message:TWMPaste);message WM_PASTE;
    procedure WMCUT (var Message:TWMCut);message WM_CUT;
    procedure WMClear (var Message:TWMClear); message WM_CLEAR;
    procedure StringsChange(Sender:TObject);
    procedure CreateExecute;
    procedure SetStrings (NewStrings:TStrings);
    procedure SetConfig(const NewConfig:TCodeEditorConfig);
    procedure SetFirstChar(Value:Integer);
    procedure SetFirstLine(Value:Integer);
    procedure IncFirstLine(Value:Integer=1);
    procedure DecFirstLine(Value:Integer=1);
    procedure IncCurY(var CurY:Integer;Value:Integer=1);
    procedure DecCurY(var CurY:Integer;Value:Integer=1);
    procedure CheckPushedCursor(const AnsichtVerschieben:Boolean=true);
    procedure CheckCurPos(var X,Y:Integer);
    procedure CheckAnsicht(const X,Y:Integer);//Sorgt dafr das die Angebene Stelle zu sehen ist.
    procedure UpdateVertScrollbar;
    procedure UpdateQuelltext;
    function VisibleLines:Integer;
    class function LTest(const S:String;Anfang:Integer=0):String;
  protected
    procedure CreateParams(var Params: TCreateParams);override;
    procedure Paint;override;
    procedure KeyPress(var Key: Char); override;
    procedure Click; override;
    procedure Resize; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    function PosRelativetoSelection(X,Y:Integer):ShortInt;
    procedure CurtoPixel(var X,Y:Integer);
    procedure PixeltoCur(var X,Y:Integer);
    procedure PixeltoChar(var X,Y:Integer);
    function LoadFromClipboard(Strings: TStrings):Boolean;
    function SavetoClipboard(Strings: TStrings):Boolean;
    function DoMouseWheel(Shift: TShiftState; WheelDelta: Integer; MousePos: TPoint): Boolean;override;
    function ZeileBeiPixel(Pixel_Y:Integer):Integer;
    procedure GetWikiCode(const Quelltext:TQuelltext;const Quelle,Ziel:TStrings);
    procedure GetXMLCode(const Quelltext:TQuelltext;const Quelle,Ziel:TStrings);
    //procedure SetCodeFont(NewCodeFont:TCodeFont);
  public
    QuelltextKlasse:TQuelltextKlasse;
    LoadedFile:String;
    NoUndo:Boolean; //Sollen Undo-Eintrge erstellt werden?
    Quelltext:TQuelltext;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    Constructor Create(AOwner:TComponent); override;
    Constructor CreateC(AOwner:TComponent;AConfig:TCodeEditorConfig;const Datei:String='');
    Destructor Destroy; override;
    procedure WortErmittlung(Y:Integer);
    function DeleteSelection:Boolean;
    function GetSelectionStrings(Strings:TStrings):Boolean;
    procedure InsertStrings(Strings:TStrings);
    procedure PasteFromClipboard;
    procedure CopytoClipboard;
    procedure UpdateHintPosition;
    property UndoList:TUndoList read FUndoList;
    property RepeatList:TRepeatList read FRepeatList;
    procedure SetCursorPos(X,Y:Integer);
    procedure ShowNewCurPos;
    procedure GetSelection(out SelBeginX,SelBeginY:Integer;out SelEndX,SelEndY:Integer);
    procedure Load(const Datei:String);

    //Fr das Menu
    procedure Menu_SelectAll;
    procedure Menu_IncreaseTabs;
    procedure Menu_DecreaseTabs;
    procedure Menu_Export_Wiki(const FileName:String);
    procedure Menu_CopyAs_WikiCode;
    procedure Menu_Export_XML(const FileName:String);
    procedure Menu_CopyAs_XMLCode;
    procedure Menu_GoTo(X,Y:Integer;NeuZeichnen:Boolean=true);
    procedure Menu_SelectLine(Y,X,Length:Integer;NeuZeichnen:Boolean=true);
    procedure Menu_ReplaceText(Y,X,Size:Integer;RText:String;NeuZeichnen:Boolean=true);
    procedure Menu_SetQuelltextTyp(Typ:TQuelltextKlasse);
    procedure Menu_Extra_XML_Umlaute_ersetzen;
  published
    property OnFunctionHint:TOnFunctionHint read FOnFunctionHint write FOnFunctionHint;
    property Lines:TStrings read FStrings write SetStrings;
    //property CodeFont:TCodeFont read FCodeFont write SetCodeFont;
    property FirstLine: Integer read FFirstLine write SetFirstLine;
    property FirstChar: Integer read FFirstChar write SetFirstChar;
    property Config:TCodeEditorConfig read FConfig write SetConfig;
    property Align;
  end;

  TOnCurPosChange=procedure(Sender:TCodeEditor;X,Y:Integer) of object;

  TCodeEditorConfig = class(TComponent)
  private
    FOnCurPosChange:TOnCurPosChange;
    FFileName:TFileName;
    FHintergrund:TColor;
    FHintergrund_Markiert:TColor;
    FMarkierterText:TColor;
    FLeerzeichenProTap:Byte;
    FSchlaueTabs:Boolean;
    procedure SetHintergrund(const NewColor:TColor);
    procedure DrawCodeEditors;
    procedure Login(const CodeEditor:TCodeEditor);
    procedure Logout(const CodeEditor:TCodeEditor);
    //function GetOrder(ShortCut:TShortCut):Integer;
    procedure NewCurPos(const Sender:TCodeEditor;const X,Y:Integer);
  protected
    CodeEditorNr:array of TCodeEditor;
    function IndexOf(const CodeEditor:TCodeEditor):Integer;
  public
    //ShortCutNr:Array[1..HighOrder] of TShortCut;
    ColorNr:TCIColors;
    StyleNr:TCIStyles;
    constructor Create(AOwner:TComponent);override;
    function LoadFromFile(FileName:String):Boolean;
    procedure ReadState(Reader: TReader); override;
    procedure LoadFromStrings(Strings:TStrings;ErlaubteReaktion:TLoadConifgReaktion=[LCR_NeuZeichnen,LCR_UpdateQuelltext]);
    function GetQuelltextTyp(const Datei:String):TQuelltextKlasse;
    property SchlaueTabs:Boolean read FSchlaueTabs write FSchlaueTabs;
    property LeerzeichenProTap:Byte read FLeerzeichenProTap write FLeerzeichenProTap;
    property MarkierterText:TColor read FMarkierterText write FMarkierterText;
    property Hintergrund_Markiert:TColor read FHintergrund_Markiert write FHintergrund_Markiert;
    property Hintergrund:TColor read FHintergrund write SetHintergrund;
  published
    property FileName:TFileName read FFileName write FFileName;
    property OnCurPosChange:TOnCurPosChange read FOnCurPosChange write FOnCurPosChange;
  end;

  TAktion=class
  private
    FOwner:TAktionsListe;
  public
    function Execute:TAktion;virtual;abstract;
    property Owner:TAktionsListe read FOwner;
  end;


  TAktionsListe=class
  private
    FSize:Integer;
    FCanUndo:Boolean;
  protected
    procedure Add(const Aktion:TAktion);virtual;
  public
    CodeEditor:TCodeEditor;
    OtherList:TAktionsListe;
    Nr:Array of TAktion;
    Constructor Create(const Owner:TCodeEditor;const NewSize:Integer);
    Destructor Destroy;override;
    procedure New(const NeueAktion:TAktion;ClearOtherList:Boolean=true);
    procedure Undo;
    procedure Clear;
    property Size:Integer read FSize;
    property CanUndo:Boolean read FCanUndo;
  end;

  TUndoList=class(TAktionsListe)
  protected
    procedure Add(const Aktion:TAktion);override;
  end;

  TRepeatList=class(TAktionsListe);


procedure Register;

implementation

const
  Maximale_Anzahl_an_Aktionen=100;
  Zeilenabstand=17;//15+2
  MaxZeichenAnzahlproZeile=300;
  ButtonSize=10;

type
//Deklaration der Aktionen


  TAktion_Delete=class(TAktion)
  private
    SelBegin:TPoint;
    SelEnd:TPoint;
  public
    function Execute:TAktion;override;
    Constructor Create(SelBeginX,SelBeginY,SelEndX,SelEndY:Integer);
  end;

  TAktion_Insert=class(TAktion)
  private
    FPos:TPoint;
    FStrings:TStrings;
  public
    function Execute:TAktion;override;
    constructor Create(DeleteStrings:TStrings;PosX,PosY:Integer);
    Destructor Destroy; override;
  end;



  TAktion_ZeichenEntfernen=class(TAktion)
  public
    Position:TPoint;
    Length:Integer;
    constructor Create(PosX,PosY:Integer;Anzahl:Integer=1);
    function Execute:TAktion;override;
    function TryAdd(Zeichen:TAktion_ZeichenEntfernen):Boolean;
  end;

  TAktion_ZeichenEingabe=class(TAktion)
  public
    Position:TPoint;
    Text:String;
    constructor Create(Zeichen:String;PosX,PosY:Integer);
    function Execute:TAktion;override;
    function TryAdd(Zeichen:TAktion_ZeichenEingabe):Boolean;
  end;

  TAktion_Return=class(TAktion)
  private
    FPos:TPoint;
  public
    function Execute:TAktion;override;
    constructor Create(PosX,PosY:Integer);
  end;

  TAktion_DeleteReturn=class(TAktion)
  private
    FPos:TPoint;
    FLAnzahl:Integer;
  public
    function Execute:TAktion;override;
    constructor Create(const XPos,YPos,LeerzeichenAnzahl:Integer);
  end;


  TAktionen=class(TAktion)
  private
    Anzahl:Integer;
  public
    Child:array of TAktion;
    function Execute:TAktion;override;
    constructor Create(const Size:Integer);
    destructor Destroy;override;
  end;



Destructor TCodeEditor.Destroy;
begin
 FStrings.Free;
 //FCodeFont.Free;
 FHintForm.Free;
 Quelltext.free;
 UndoList.free;
 RepeatList.free;
 //for I:= 0 to High(FontNr) do    FontNr[I].free;
 inherited;

end;

procedure TCodeEditor.CreateExecute;
begin
 FStrings:= TStringList.Create;
 TStringList(FStrings).OnChange := StringsChange;
 FUpdatingStrings := true;
 //FStrings.Add('');
 FUpdatingStrings := false;
 //FCodeFont:=TCodeFont.Create;
 //FCodeFont.OnChange := Refresh;
 FHintForm := THintForm.CreateNew(self);
 DoubleBuffered := true;
 TabStop := true;
 FAbstandzumRand:= 20;//Pixel
 ControlStyle := ControlStyle - [csAcceptsControls, csSetCaption] + [csFramed, csOpaque, csMenuEvents];
 Quelltext := QuelltextKlasse.Create(Lines);
 FUndoList := TUndoList.create(self,Maximale_Anzahl_an_Aktionen);
 FRepeatList := TRepeatList.Create(self,Maximale_Anzahl_an_Aktionen);
 FUndoList.OtherList := FRepeatList;
 FRepeatList.OtherList := FUndoList;
 Cursor := crIBeam;
 Canvas.Font.Name := 'Courier New';
 Canvas.Font.Pitch := fpFixed;
 Canvas.font.Size := (Zeilenabstand*3) div 4 -2;
end;

Constructor TCodeEditor.Create(AOwner:TComponent);
begin
 inherited;
 FConfig := TCodeEditorConfig.Create(self);
 QuelltextKlasse := TXMLQuelltext;
 CreateExecute;
end;

Constructor TCodeEditor.CreateC(AOwner:TComponent;AConfig:TCodeEditorConfig;const Datei:String='');
begin
 inherited Create(AOwner);
 if Assigned(AConfig) then FConfig := AConfig else  FConfig := TCodeEditorConfig.Create(self);

 Align := alClient;
 QuelltextKlasse := Config.GetQuelltextTyp(Datei);
 CreateExecute;
end;
procedure TCodeEditor.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  Params.Style := Params.Style or WS_HSCROLL or WS_VSCROLL;
end;

//Link: Create


procedure TCodeEditor.WMCreate(var Message: TWMCreate);
var
  ScrollInfo:tagScrollInfo;
begin
 inherited;

 //---------------------------------------
 FTimerID := 111;
 SetTimer(Handle,FTimerID,100,nil);

 ScrollInfo.fMask := SIF_PAGE or SIF_RANGE or SIF_DISABLENOSCROLL;

 ScrollInfo.nPage := 50	;
 ScrollInfo.nMin := 0;
 ScrollInfo.nMax := 50+ MaxZeichenAnzahlproZeile-1;
 SetScrollInfo(Handle,SB_HORZ,ScrollInfo,true);

 ScrollInfo.nPage := 49	;
 ScrollInfo.nMin := 0;
 ScrollInfo.nMax := 50;
 SetScrollInfo(Handle,SB_VERT,ScrollInfo,true);

 SetFirstLine(0);
 SetFirstChar(0);


 (*ShowScrollBar(Handle,SB_BOTH,True); berflssig durch CreateParams*)

 //SetScrollRange(Handle,SB_VERT,0,1,true);
 //SetScrollRange(Handle,SB_HORZ,0,MaxZeichenAnzahlproZeile-1,true);
end;




procedure TCodeEditor.StringsChange(Sender:TObject);
begin
 if FUpdatingStrings then exit;
 UpdateVertScrollbar;
 UpdateQuelltext;
end;

procedure TCodeEditor.WMTimer(var Message :TWMTimer);
const
  SpeedX=10;     //Link: WMTIMER
begin
 inherited;
 Message.result := 0;
 //Blinken
 Inc(FBlinkZeit,100);

 if FBlinkZeit >= getCaretBlinkTime then
 begin
  FTextCursorVisible := not FTextCursorVisible;
  FBlinkZeit := 0;
  //Neu zeichnen
  Refresh;
 end;

 if FMouseDown then CheckPushedCursor;

 //Hinweis verschieben
 UpdateHintPosition; //Wann?? Stand vorher vor Refresh;
end;
procedure TCodeEditor.GetXMLCode(const Quelltext:TQuelltext;const Quelle,Ziel:TStrings);
  function Leerzeichen(const Anzahl:Integer):String;
  var
    I:Integer;
  begin
    SetLength(result,Anzahl);
    for I := 1 to length(result) do result[I] := ' ';
  end;
var
  AFontIndex:Integer;
  function AnfangvomWort(Wort:TWort):String;
  begin
    if (AFontIndex <> Wort.FontIndex) and (Wort.FontIndex <> CI_Normal) then
    begin
      case Wort.FontIndex of
      CI_Compiler:
        Result := '<' + CIName[CI_Compiler] +' type="compiler">';
      else
        Result := '<' + CIName[Wort.FontIndex] + '>';
      end;
    end
    else
    Result := '';
    AFontIndex := Wort.FontIndex;
  end;
  function EndevomWort(const Next_Wort:TWort):String;
  begin
    Result := '';
    if AFontIndex <> Next_Wort.FontIndex then //Es gab eine Farbnderung
    if AFontIndex <> CI_Normal then //Und es gibt was zu schlieen
    begin
      result := '</' + CIName[AFontIndex] + '>';
    end;
  end;
  function XMLText(Text:String):String;
  var
    I:Integer;
    Max:Integer;
  begin
    Max := length(Text);
    Result := Text;
    for I := 1 to Max do
    begin
      if (Byte(Text[I]) < 40) or (Byte(Text[I]) > 126) or (Text[I] = '<') or (Text[I] = '>') then
      begin//Ungltige Zeichen von diesem Typ ersetzen
       Result := StringReplace(Result,Text[I],'&#' + IntToStr(Byte(Text[I])) + ';',[rfReplaceAll]);
       Text := StringReplace(Text,Text[I],'H',[rfReplaceAll]);//Im Ausgangstext alle Vorkommen lschen.
      end;
    end;
  end;
var
  X,Y:Integer;
  S:String;
  YMax,XMax:Integer;
  Level:Integer;
  LevelSize:array of Integer;
  high_LevelSize:Integer;
  LevelSizeTotal:Integer;
  LeerzeichenAnzahl:Integer;
  Schlusswort:TWort;
  ZeilenumbruchNachtragen:Boolean;
  LetzteRichtigeZeile:Integer;
  function XMLTapLeerzeichen:String;
  begin
    Result := Leerzeichen((Level+1)*StdLeerzeichenProTap);
  end;
begin  {GetXMLCode(const Quelltext:TQuelltext;const Quelle,Ziel:TStrings); }
  if Ziel = nil then raise Exception.Create('GetXMLCode: Ziel ist nicht vorhanden');
  high_LevelSize := 50;
  SetLength(LevelSize,high_LevelSize+1); //Erstmal auf 50 setzen, um nacher nicht erhhen zu mssen.
  Ziel.Clear;
  YMax := Quelle.count-1;
  Level := 0;
  LevelSizeTotal := 0;
  Ziel.Add('<code>');
  ZeilenumbruchNachtragen := false;
  LetzteRichtigeZeile := 0;
  for Y := 0 to YMax do
  with Quelltext do
  begin
    if Zeilen[Y] = nil then
    begin
      Ziel.Add(XMLTapLeerzeichen + '<br />');
      Continue;
    end;
    XMax := high(Zeilen[Y])-1;
    LeerzeichenAnzahl := Zeilen[Y][0].Anfang-1;
    if LeerzeichenAnzahl > LevelSizeTotal then
    begin
      if LeerzeichenAnzahl = LevelSizeTotal + StdLeerzeichenProTap then
       Ziel.Add(XMLTapLeerzeichen + '<level>')
      else
       Ziel.Add(XMLTapLeerzeichen + '<level size="' + IntToStr(Zeilen[Y][0].Anfang-LevelSizeTotal-1) + '">');
      Inc(Level);
      //Fr Platz zum Speichern sorgen.
      if Level > high_LevelSize then
      begin
        Inc(high_LevelSize,50);
        SetLength(LevelSize,high_LevelSize+1);
      end;
      LevelSize[Level] := LeerzeichenAnzahl-LevelSizeTotal;
      LevelSizeTotal := LeerzeichenAnzahl;
      ZeilenumbruchNachtragen := false;
    end;
    while LeerzeichenAnzahl < LevelSizeTotal do
    begin
      Dec(LevelSizeTotal,LevelSize[Level]);
      Dec(Level);
      Ziel.Add(XMLTapLeerzeichen + '</level>');
      ZeilenumbruchNachtragen := false;
    end;
    if ZeilenumbruchNachtragen then
    begin
      S := Ziel[LetzteRichtigeZeile];
      Ziel[LetzteRichtigeZeile] := Ziel[LetzteRichtigeZeile] + '<br />'
    end;

    AFontIndex := CI_Normal;
    S  := XMLTapLeerzeichen;
    for X := 0 to XMax do
    begin
      S := S + AnfangvomWort(Zeilen[Y][X]);
      S := S + XMLText(Zeilen[Y][X].Text);
      S := S + EndevomWort(Zeilen[Y][X+1]);
      S := S + LeerZeichen(Zeilen[Y][X+1].Anfang - (length(Zeilen[Y][X].Text) + Zeilen[Y][X].Anfang));
    end;
    S := S + AnfangvomWort(Zeilen[Y][XMax+1]);
    S := S + XMLText(Zeilen[Y][XMax+1].Text);

    Schlusswort.Anfang := Length(Quelle.Strings[Y]);
    Schlusswort.Text := '';
    Schlusswort.FontIndex := 0;
    S := S + EndevomWort(Schlusswort);
    LetzteRichtigeZeile := Ziel.Add(S);
    ZeilenumbruchNachtragen := true;
  end;
  while Level > 0 do
  begin
      Dec(Level);
      Ziel.Add(XMLTapLeerzeichen + '</level>');
  end;
  Ziel.Add('</code>');
end;


procedure TCodeEditor.GetWikiCode(const Quelltext:TQuelltext;const Quelle,Ziel:TStrings);
  function Leerzeichen(const Anzahl:Integer):String;
  var
    I:Integer;
  begin
    SetLength(result,Anzahl);
    for I := 1 to length(result) do result[I] := ' ';
  end;
const
  StandartFarbe:TColor=clWindowText;
var
  S:String;
  YMax,XMax:Integer;
  X,Y:Integer;
  AStyle:TFontStyles;
  AColor:TColor;
  Wort_Color:TColor;
  Wort_ist_fett:Boolean;
  Wort_ist_kursiv:Boolean;
  SchlussWort:TWort;
  function ColorToInt(const Color:TColor):Integer;
  type
    TBGRColor=record B,G,R,U :Byte;end;
  var
    RGBColor:TBGRColor;
    Rot:Byte;
  begin
    RGBColor :=TBGRColor(ColorToRGB(Wort_Color));
    Rot := RGBColor.R;
    RGBColor.R := RGBColor.B;
    RGBColor.B := Rot;
    result := LongInt(RGBColor);
  end;
  function AnfangvonWort(Wort:TWort):String;
  begin
    result := '';
    Wort_Color := StdColorNr[Wort.FontIndex];
    Wort_ist_fett:= fsBold in StdStyleNr[Wort.FontIndex];
    Wort_ist_kursiv:= fsItalic in StdStyleNr[Wort.FontIndex];

    if (AColor <> Wort_Color) and (ColorToRGB(Wort_Color) <> ColorToRGB(StandartFarbe)) then
    begin
      result := result +'<font color="#'+IntToHex(ColorToInt(Wort_Color),6)+'">';
    end;

    if Wort_ist_kursiv and not(fsItalic in AStyle) then
    begin
      result := result + '<i>';
    end;

    if Wort_ist_fett and not(fsbold in AStyle) then
    begin
      result := result + '<b>';
    end;

    AStyle:= StdStyleNr[Wort.FontIndex];
    AColor:= StdColorNr[Wort.FontIndex];
  end;
  function EndevomWort(const Next_Wort:TWort):String;
  var
    Close_Color:Boolean;
    Close_Bold:Boolean;
    Close_Italic:Boolean;
  begin
    result := '';
    Wort_Color := StdColorNr[Next_Wort.FontIndex];
    Wort_ist_fett:= fsBold in StdStyleNr[Next_Wort.FontIndex];
    Wort_ist_kursiv:= fsItalic in StdStyleNr[Next_Wort.FontIndex];

    Close_Color :=false;
    Close_Bold  :=false;
    Close_Italic:=false;
    if AColor <> Wort_Color then
    begin//Es gab eine Farbnderung
      if ColorToRGB(AColor) <> ColorToRGB(StandartFarbe) then Close_Color :=true;
      Close_Bold  :=true;
      Close_Italic:=true;
    end;
    if (fsItalic in AStyle) <> Wort_ist_kursiv  then
    begin //Es gab eine Kursiv-nderung
      if (fsItalic in AStyle) then Close_Italic:=true;
      Close_Bold  :=True;
    end;
    if (fsBold in AStyle) <> Wort_ist_fett  then
    begin //Es gab eine Fett-nderung
      if (fsBold in AStyle) then  Close_Bold  :=True;
    end;

    if Close_Bold  and (fsBold in AStyle)then
    begin
      result := result + '</b>';
      AStyle:= AStyle - [fsBold]
    end;
    if Close_Italic and (fsItalic in AStyle) then
    begin
      result := result + '</i>';
      AStyle:= AStyle - [fsItalic]
    end;
    if Close_Color then
    begin
      result := result + '</font>';
      AColor := StandartFarbe;
    end;
  end;
{procedrue GetWikiCode(const Ziel:TStrings);}
begin
  if Ziel = nil then raise Exception.Create('GetWikiCode: Ziel ist nicht vorhanden');
  Ziel.Clear;
  YMax := Quelle.count-1;
  AStyle:= [];
  AColor:= StandartFarbe;
  for Y := 0 to YMax do
   with Quelltext do
   begin
     XMax := high(Zeilen[Y])-1;
     if XMax = -2 then
     begin
       Ziel.Add(' <br />');
       Continue;
     end;
     X := 0;
     S := LeerZeichen(Zeilen[Y][X].Anfang);
     for X := 0 to XMax do
     begin
       S := S + AnfangvonWort(Zeilen[Y][X]);
       S := S + Zeilen[Y][X].text;
       S := S + EndevomWort(Zeilen[Y][X+1]);
       S := S + LeerZeichen(Zeilen[Y][X+1].Anfang - (length(Zeilen[Y][X].Text) + Zeilen[Y][X].Anfang));
     end;
     S := S + AnfangvonWort(Zeilen[Y][XMax+1]);
     S := S + Zeilen[Y][XMax+1].text;

     Schlusswort.Anfang := Length(Quelle.Strings[Y]);
     Schlusswort.Text := '';
     Schlusswort.FontIndex := 0;
     S := S + EndevomWort(Schlusswort);

     Ziel.Add(S);
   end;
end;

//Link: Wiki Export

procedure TCodeEditor.Menu_CopyAs_WikiCode;
var
  AusgabeStrings:TStrings;
  EingabeStrings:TStrings;
  EingabeQuelltext:TQuelltext;
begin
  AusgabeStrings := TStringList.Create;
  try
    EingabeStrings := TStringList.Create;
    try
    if GetSelectionStrings(EingabeStrings) then //Falls etwas Markiert ist
    begin   //Markierten Bereich highlighten
      EingabeQuelltext := QuelltextKlasse.Create(EingabeStrings);
      try
        GetWikiCode(EingabeQuelltext,EingabeStrings,AusgabeStrings)
      finally
        EingabeQuelltext.Free;
      end;
    end
    else
     GetWikiCode(Quelltext,FStrings,AusgabeStrings);
    finally
      EingabeStrings.Free;
    end;
    SavetoClipboard(AusgabeStrings)
  finally
    AusgabeStrings.Free;
  end;
end;
procedure TCodeEditor.Menu_CopyAs_XMLCode;
var
  AusgabeStrings:TStrings;
  EingabeStrings:TStrings;
  EingabeQuelltext:TQuelltext;
begin
  AusgabeStrings := TStringList.Create;
  try
    EingabeStrings := TStringList.Create;
    try
    if GetSelectionStrings(EingabeStrings) then //Falls etwas Markiert ist
    begin   //Markierten Bereich highlighten
      EingabeQuelltext := QuelltextKlasse.Create(EingabeStrings);
      try
        GetXMLCode(EingabeQuelltext,EingabeStrings,AusgabeStrings)
      finally
        EingabeQuelltext.Free;
      end;
    end
    else
     GetXMLCode(Quelltext,FStrings,AusgabeStrings);
    finally
      EingabeStrings.Free;
    end;
    SavetoClipboard(AusgabeStrings)
  finally
    AusgabeStrings.Free;
  end;
end;

procedure TCodeEditor.Menu_Export_XML(const FileName:String);
var
  AusgabeStrings:TStrings;
begin
  AusgabeStrings := TStringList.Create;
  try
    GetXMLCode(Quelltext,FStrings,AusgabeStrings);
    AusgabeStrings.SaveToFile(FileName);
  finally
    AusgabeStrings.Free;
  end;
end;

procedure TCodeEditor.Menu_Export_Wiki(const FileName:String);
var
  AusgabeStrings:TStrings;
begin
  AusgabeStrings := TStringList.Create;
  try
    GetWikiCode(Quelltext,FStrings,AusgabeStrings);
    AusgabeStrings.SaveToFile(FileName);
  finally
    AusgabeStrings.Free;
  end;
end;

procedure TCodeEditor.Menu_Extra_XML_Umlaute_ersetzen;
  function Umlaute_ersetzen(Text:String):String;
  var
    I:Integer;
    Max:Integer;
  begin
    Max := length(Text);
    Result := Text;
    for I := 1 to Max do
    begin
      if Byte(Text[I]) > 126  then
      begin//Ungltige Zeichen von diesem Typ ersetzen
       Result := StringReplace(Result,Text[I],'&#' + IntToStr(Byte(Text[I])) + ';',[rfReplaceAll]);
       Text := StringReplace(Text,Text[I],'H',[rfReplaceAll]);//Im Ausgangstext alle Vorkommen lschen.
      end;
    end;
  end;
var
  Y:Integer;
begin
  FUpdatingStrings := true;
  for Y := 0 to Lines.Count-1 do
   Lines[Y] := Umlaute_ersetzen(Lines[Y]);
  FUpdatingStrings := false;
  UpdateQuelltext;
  Refresh;
end;

procedure TCodeEditor.UpdateHintPosition;
var X,Y:Integer;
begin
 X := FHintCurPos.X;
 Y := FHintCurPos.Y;
 CurtoPixel(X,Y);
 FHintForm.Moveto( ClientOrigin.X+X , ClientOrigin.Y+Y);
end;

procedure TCodeEditor.WortErmittlung(Y:Integer);
begin
  {$IFDEF ERRORCHECK}
  if Y > FStrings.count-1 then  raise Exception.Create(Format('Fehler: Y (%d) >FStrings.count-1 (%d)',[Y,FStrings.count-1]));
  if Y > high(Quelltext.Zeilen) then  raise Exception.Create(Format('Fehler: Y=%d >high(Quelltext.Zeilen)=%d',[Y,high(Quelltext.Zeilen)]));
  {$ENDIF}
  Quelltext.WortErmittlung(Quelltext.Zeilen[Y],FStrings[Y]);
end;

function TCodeEditor.LoadFromClipboard(Strings: TStrings):Boolean;
var
 Text:PChar;
 H:HGLOBAL;
 I:Integer;
 LastFormat:UINT;
 Max:Integer;
 Text_available:Boolean;
begin
 result := false;
 if not OpenClipboard(Handle) then Exit;

 //Feststellen ob Text in der Zwischenablage ist.
 LastFormat := EnumClipboardFormats(0);
 Max := CountClipboardFormats-1;
 Text_available := False;
 for I := 0 to Max do
 begin
  LastFormat := EnumClipboardFormats(LastFormat);
  if (LastFormat = CF_TEXT) or (LastFormat = CF_OEMTEXT) or (LastFormat = CF_UNICODETEXT) then
  begin
    Text_available := True;
    Break;
  end;
 end;
 //Falls kein
 if not Text_available then Exit;
 H := GetClipboardData(CF_TEXT);
 Text := GlobalLock(H); //Gibt einen Zeiger auf die Globale Variable zurck
 Strings.Text := Text;
 GlobalUnlock(H); //Andere Anwendungen drfen wieder auf die globale Variable zugreifen
 CloseClipboard;
 result := True;
 
end;

function TCodeEditor.SavetoClipboard(Strings: TStrings):Boolean;
var
 Text:PChar;
 H:HGLOBAL;
 S:String;
 I:Integer;
begin
 result := false;
 if not OpenClipboard(Handle) then exit;
 S:=strings.text;
 H:=GlobalAlloc(GMEM_MOVEABLE or GMEM_DDESHARE,Length(S)+1);
 Text := GlobalLock(H);
 for I := 1 to Length(S) do
 begin
  Text^ := S[I];
  Inc(Text);
 end;
 Text^ := #0;//Letztes Zeichen beschreibt das Ende
 GlobalUnLock(H);//Fr andere Anwendugen verfgbar machen
 EmptyClipboard;//Zwischenablage leeren
 SetClipBoardData(CF_TEXT,H);
 CloseClipboard;
 result := true;
end;


procedure TCodeEditor.UpdateVertScrollbar;
var
  ScrollInfo:tagScrollInfo;
begin
(*
  if Win32Platform < 2 then //Falls noch kein NT Verwendet wird
    SetScrollRange(Handle,SB_VERT,0,49 + FStrings.Count-1,True)
  else  *)
  begin  
    ScrollInfo.fMask := SIF_PAGE or SIF_RANGE ;
    ScrollInfo.nPage := 50;
    ScrollInfo.nMin := 0;
    ScrollInfo.nMax := 49+ FStrings.Count-1;
    SetScrollInfo(Handle,SB_VERT,ScrollInfo,true);
  end;
end;

procedure TCodeEditor.WMDblClick(var Message:TWMLButtonDblClk );
var
  CurPos:TPoint;
begin
 inherited;
 if Message.XPos < FAbstandzumRand then exit;
 //CurPos:=Mouse.CursorPos;
 CurPos.X := Message.XPos;
 CurPos.Y := Message.YPos;

 PixeltoCur(CurPos.X,CurPos.Y);
 FSelX := 0;
 FSelY := CurPos.Y;
 CheckCurPos(FSelX,FSelY);

 FCurX := 0;
 FCurY := FSelY+1;
 CheckCurPos(FCurX,FCurY);
 if (FSelY = FStrings.count-1) and (FStrings.Count <> 0)then
 begin
   FCurY := FStrings.count-1;
   FCurX := length(FStrings[FCurY]);
 end;
 CheckAnsicht(FCurX,FCurY);
 FDoubleClick := true;
 Refresh;
end;

procedure TCodeEditor.CurtoPixel(var X,Y:Integer);
begin
 Y := (Y-FFirstLine) * Zeilenabstand;
 X := (X - FirstChar)* Canvas.TextWidth(' ') + FAbstandzumRand;
end;

procedure TCodeEditor.PixeltoChar(var X,Y:Integer);
begin  
  Y:= ZeileBeiPixel(Y);
  X:= round((X-FAbstandzumRand) / Canvas.TextWidth(' '))+FFirstChar;
end;
procedure TCodeEditor.PixeltoCur(var X,Y:Integer);
begin
  Y:= ZeileBeiPixel(Y);
  X:= Trunc((X-FAbstandzumRand) / Canvas.TextWidth(' '))+FFirstChar;
end;  //Hinweis Menu_GoTo benutz ebenfalls eine solche Rechnung
function TCodeEditor.DeleteSelection:Boolean;
var
  S:String;
  I:Integer;
  StringstoDelete:TStrings;//Fr das Rckgngig machen
  SelBeginX,SelBeginY:Integer;
  SelEndX,SelEndY:Integer;
begin
 GetSelection(SelBeginX,SelBeginY,SelEndX,SelEndY);

 StringstoDelete := TStringList.Create;
 if GetSelectionStrings(StringstoDelete) then
 begin
  FUndoList.New(TAktion_Insert.Create(StringstoDelete,SelBeginX,SelBeginY)) ;
  result := true;
 end
 else
 begin
  StringstoDelete.free;
  result := false;
 end;

 FCurX := SelBeginX;
 FCurY := SelBeginY;
 FSelX := SelBeginX;
 FSelY := SelBeginY;



 if FStrings.Count = 0 then FStrings.Add('');
 if SelBeginY = SelEndY then
 begin
  S:= FStrings[SelBeginY];
  Delete(S,SelBeginX+1,SelEndX-SelBeginX);
  FStrings[SelBeginY] := LTest(S,SelBeginX);
  WortErmittlung(SelBeginY);
  Quelltext.WortErkennung(SelBeginY,SelBeginY);
  exit;
 end;
 //Obere Zeile ergenzen
 S:= FStrings[SelBeginY];
 Delete(S,SelBeginX+1,length(S)-SelBeginX);
 FStrings[SelBeginY] := LTest(S + copy(FStrings[SelEndY],SelendX+1,length(FStrings[SelEndY])-SelEndX));
 //Rest Lschen
 for I:= SelEndY downto SelBeginY+1  do
 FStrings.Delete(I);
 //nderungen auch an Objekt "Quelltext" durchfhren
 Quelltext.DeleteLines(SelBeginY,SelEndY-SelBeginY);
 WortErmittlung(SelBeginY);
 Quelltext.WortErkennung(SelBeginY,SelBeginY);

end;



procedure TCodeEditor.InsertStrings(Strings:TStrings);
var
  S:String;
  I:Integer;
begin
  if (Strings = nil) or (Strings.count=0) then   exit;

  S:= FStrings[FCurY];
  while length(S) < FCurX do S := S + ' ';

  if Strings.Count = 1 then //Nur eine Zeile einzufgen
  begin
    System.Insert(Strings[0],S,FCurX+1);
    FStrings[FCurY] := LTest(S);
    FUndoList.New(TAktion_Delete.Create(FCurX,FCurY,FCurX + length(Strings[0]),FCurY));

    //nderungen auch an der Eigenschaft "Quelltext" durchfhren.
    WortErmittlung(FCurY);
    Quelltext.WortErkennung(FCurY,FCurY);

    SetCursorPos(FCurX + length(Strings[0]),FCurY);
    exit;
  end;


  FUndoList.New(TAktion_Delete.Create(FCurX,FCurY,length(Strings[Strings.count-1]),FCurY+Strings.count-1));

  //Markierungs Ende
  FStrings.Insert(FCurY+1,LTest(Strings[Strings.count-1] + copy(S,FCurX+1,length(S)-FCurX+1)) );
  //Mitte
  for I := Strings.count-2 downto 1 do 
    FStrings.Insert(FCurY+1, LTest(Strings[I]));

  //Anfang
  FStrings[FCurY]:= LTest(copy(S,1,FCurX) + Strings[0]);

  //nderungen auch an der Eigenschaft "Quelltext" durchfhren.
  Quelltext.InsertLines(FCurY,Strings.Count-1);
  for I := FCurY to FCurY + Strings.Count-1 do WortErmittlung(I);
  Quelltext.WortErkennung(FCurY,FCurY + Strings.Count-1);
  
  SetCursorPos(length(Strings[Strings.count-1]),FCurY + Strings.count-1);
end;

function TCodeEditor.GetSelectionStrings(Strings:TStrings):Boolean;
var
  S:String;
  I:Integer;
  SelBeginX,SelBeginY:Integer;
  SelEndX,SelEndY:Integer;
begin
 result := false;
 if (Strings = nil) then exit;
 if FStrings.count = 0 then exit;
 CheckCurPos(FCurX,FCurY);

 //Die Markierungs Daten erhalten
 GetSelection(SelBeginX,SelBeginY,SelEndX,SelEndY);

 //berprfen ob berhaupt etwas markiert ist
 if (SelBeginX = SelEndX) and (SelBeginY = SelEndY) then Exit;

 //Nun liegt sicher eine Markierung vor
 result  := true;
 Strings.Clear;

 //Markierung betrifft nur eine Zeile
 if SelBeginY = SelEndY then
 begin
  S := FStrings[SelBeginY];
  Strings.Add(copy(S,SelBeginX+1,SelEndX-SelBeginX));
  exit;
 end;


 //Markerungs Anfang
 S := FStrings[SelbeginY];
 Strings.Add(copy(S,SelBeginX+1,length(S)-SelBeginX+1));

 //Markierungs Mitte
 for I:= SelBeginY+1 to SelEndY-1 do
 Strings.Add(FStrings[I]);

 //Markierungs Ende
 S := FStrings[SelEndY];
 Strings.Add(copy(S,1,SelEndX));
end;

procedure TCodeEditor.PasteFromClipboard;
var Strings:TStrings;   //Link: PasteFromClipBoard
begin
 FUpdatingStrings := true;

 Strings := TStringList.Create;
 try
  DeleteSelection;
  LoadFromClipBoard(Strings);
  InsertStrings(Strings);
 finally
  Strings.Free;
 end;
 FUpdatingStrings := false;

end;

procedure TCodeEditor.CopytoClipboard;
var Strings:TStrings;
begin
 Strings := TStringList.Create;
 try
  if GetSelectionStrings(Strings) then SavetoClipboard(Strings);
 finally
  Strings.Free;
 end;
end;

procedure TCodeEditor.GetSelection(out SelBeginX,SelBeginY:Integer;out SelEndX,SelEndY:Integer);
var
 CuristSelBegin:Boolean;
begin
 if FCurY = FSelY then
 CuristSelBegin:= FCurX < FSelX
 else
 CuristSelBegin:= FCurY < FSelY;
 if CuristSelBegin then
 begin
  SelBeginX :=FCurX;
  SelBeginY :=FCurY;
  SelEndX := FSelX;
  SelEndY := FSelY;
 end
 else
 begin
  SelBeginX :=FSelX;
  SelBeginY :=FSelY;
  SelEndX := FCurX;
  SelEndY := FCurY;
 end;
   //Fr Markierungen:
 (*
 if SelEndY < FStrings.count then
 if SelEndX > length(FStrings[SelEndY]) then SelEndX := length(FStrings[SelEndY]);
 *)
end;

function TCodeEditor.PosRelativetoSelection(X,Y:Integer):ShortInt;
const
 danach=1;
 davor=-1;
 dazwischen=0;
var
 SelBeginX,SelBeginY:Integer;
 SelEndX,SelEndY:Integer;
begin
 GetSelection(SelBeginX,SelBeginY,SelEndX,SelEndY);


 //Nur eine Zeile markiert
 if (SelBeginY = Y) and (SelEndY =Y) then
 begin
  if X >= SelEndX then result := danach
  else
  if X < SelBeginX then result := davor
  else
  result := dazwischen;
  exit;
 end;

 //Mitte
 if  ((SelBeginY < Y) and (Y < SelEndY)) then
 begin
  result := dazwischen;
  exit;
 end;

 //Markierung in der oberen Zeile
 if (SelBeginY = Y) then
 begin
  if X >= SelBeginX then result := dazwischen else result := davor;
  exit;
 end;
 if (SelEndY = Y) then
 begin
  if X < SelEndX then result := dazwischen else result := danach;
  exit;
 end;

 if Y > SelBeginY then result := danach
 else
 result := davor;
end;

procedure TCodeEditor.WMGetDlgCode(var Message: TWMGetDlgCode);
begin
 Message.Result := DLGC_WANTALLKEYS or DLGC_WANTARROWS or DLGC_WANTTAB;
end;

procedure TCodeEditor.WMKeyDown(var Message:TWMKey); //Link: WMKeyDown
 procedure GoToWord(Next:Boolean);
 var
   I:Integer;

 begin
   {$IFDEF ERRORMESSAGE}
   if FCurY > high(Quelltext.Zeilen) then raise Exception.Create(Format('Zeile %d gibt es nicht(Max: %d)',[FCurX,high(Quelltext.Zeilen)]));
   {$ENDIF}
   if Next then
   begin
    for I := 0 to high(Quelltext.Zeilen[FCurY]) do
    if  Quelltext.Zeilen[FCurY,I].Anfang+length(Quelltext.Zeilen[FCurY,I].Text)-1 > FCurX then
    begin
      FCurX := Quelltext.Zeilen[FCurY,I].Anfang+length(Quelltext.Zeilen[FCurY,I].Text)-1;
      exit;
    end;
    Inc(FCurY);
    if FCurY < FStrings.count then FCurX := 0;
   end
   else
   begin
    for I := high(Quelltext.Zeilen[FCurY]) downto 0 do
    if  Quelltext.Zeilen[FCurY,I].Anfang-1 < FCurX then
    begin
      FCurX := Quelltext.Zeilen[FCurY,I].Anfang-1;
      exit;
    end;
    Dec(FCurY);
    if (FCurY < 0) then exit;

    if length(Quelltext.Zeilen[FCurY]) = 0 then
     FCurX := 0
    else
     with Quelltext.Zeilen[FCurY,high(Quelltext.Zeilen[FCurY])] do FCurX := Anfang + length(Text);


   end
 end;
var
  ShiftState:TShiftState;
  S:String;
const
  U = 85;
  I = 73;
begin
 FBlinkZeit := 0;
 FTextCursorVisible := true;
 ShiftState := [];
 if (GetKeyState(VK_SHIFT) and 128) <> 0 then ShiftState := ShiftState + [ssShift];
 if (GetKeyState(VK_MENU) and 128) <> 0 then ShiftState := ShiftState + [ssAlt];
 if (GetKeyState(VK_CONTROL) and 128) <> 0 then ShiftState := ShiftState + [ssCtrl];

 //Pfeiltasten prfen
 case Message.CharCode of VK_UP,VK_DOWN,VK_RIGHT,VK_LEFT:
  begin
   if  ssCTRL in ShiftState then
   case Message.CharCode of
    VK_UP:  DecCurY(FCurY);
    VK_DOWN : IncCurY(FCurY);
    VK_LEFT : GoToWord(false);
    VK_RIGHT:  GoToWord(true);
   end
   else//Kein Strg
   case Message.CharCode of
    VK_UP:  DecCurY(FCurY);
    VK_DOWN : IncCurY(FCurY);
    VK_LEFT : Dec(FCurX);
    VK_RIGHT:  Inc(FCurX);
   end;
   CheckCurPos(FCurX,FCurY);
   CheckAnsicht(FCurX,FCurY);
   if not  (ssShift in ShiftState) then
   begin
     FSelX := FCurX;
     FSelY := FCurY;

   end;
   Message.Result := 0;
   Refresh;
   ShowNewCurPos;
  end;
 end;

 case Message.CharCode of

  VK_PRIOR://Bild Auf
   begin
     DecCurY(FCurY,VisibleLines);
     DecFirstLine(VisibleLines);
     CheckCurPos(FCurX,FCurY);
     if not(ssShift in ShiftState) then FSelY := FCurY;
     message.result := 0;
     Refresh;
     ShowNewCurPos;
   end;
  VK_NEXT://Bild ab
   begin
     IncCurY(FCurY,VisibleLines);
     IncFirstLine(VisibleLines);
     CheckCurPos(FCurX,FCurY);
     if not(ssShift in ShiftState) then FSelY := FCurY;
     message.result := 0;
     Refresh;
     ShowNewCurPos;
   end;

  VK_HOME://Pos1
  begin
   if ssCTRL in ShiftState then
   begin //Ganz an den Anfang
     FirstLine:=0;
     FirstChar:=0;
     FCurX := 0;
     FCurY := 0;
     CheckCurPos(FCurX,FCurY);
     if not(ssShift in ShiftState) then
     begin
       FSelX := FCurX;
       FSelY := FCurY;
     end;
   end
   else
   begin//Anfang der Zeile
     FirstChar:= 0;
     FCurX := 0;
     CheckCurPos(FCurX,FCurY);
     if not(ssShift in ShiftState) then FSelX := FCurX;
   end;
   message.result := 0;
   Refresh;
   ShowNewCurPos;
  end;
  VK_END://Ende
  if FStrings.Count <> 0 then
  begin
   if ssCTRL in ShiftState then
   begin
     FirstLine:=Lines.Count - (FLastChar-FFirstChar) ;
     FCurY := Lines.Count-1;
     FirstChar:= length(Lines[FCurY]) - (FLastChar-FFirstChar);
     FCurX := length(Lines[FCurY]);
     CheckCurPos(FCurX,FCurY);
     if not(ssShift in ShiftState) then
     begin
       FSelX := FCurX;
       FSelY := FCurY;
     end;
   end
   else
   begin
    FirstChar:= length(Lines[FCurY]) - (FLastChar-FFirstChar)+1;
    FCurX := length(Lines[FCurY]);
    CheckCurPos(FCurX,FCurY);
    if not(ssShift in ShiftState) then FSelX := FCurX;
   end;
   message.result := 0;
   Refresh;
   ShowNewCurPos;
  end;
  VK_DELETE:
  if FStrings.Count <> 0 then
  begin
   FUpdatingStrings := true;
   if not((FSelX = FCurX) and (FSelY = FCurY)) then
   begin
    DeleteSelection;
   end
   else
   begin
    if (FCurX < length(Lines[FCurY])) then
    begin //Zeichen Lschen
     S:= FStrings[FCurY];
     UndoList.New(TAktion_ZeichenEingabe.Create(S[FCurX+1],FCurX+1,FCurY));
     Delete(S, FCurX+1, 1);
     FStrings[FCurY] := LTest(S);
     //nderungen auch an Objekt Quelltext durchfhren.
     Wortermittlung(FCurY);
     Quelltext.WortErkennung(FCurY,FCurY);
    end
    else
    if (FCurY+1 < Lines.count) then
    begin //Zeilenumbruch lschen.
     //FCurX := length(FStrings[FCurY-1]);
     S := FStrings[FCurY];
     while length(S) < FCurX do S := S + ' ';
     FStrings[FCurY] := S+FStrings[FCurY+1];
     FStrings.Delete(FCurY+1);
     UndoList.New(TAktion_Return.Create(FCurX,FCurY));
     //nderungen auch an Objekt Quelltext durchfhren.
     Quelltext.DeleteLines(FCurY+1,1);
     WortErmittlung(FCurY);
     Quelltext.WortErkennung(FCurY,FCurY);
    end
   end;
   FUpdatingStrings := false;
   Message.result := 0;
   Refresh;
  end;
 end;


 (*
   //ShortCuts abfragen
 Order := Config.GetOrder(ShortCut(Message.CharCode,ShiftState));
 if Order <> Order_DoNothing then
 begin
  case Order of

  end;
  Message.Result := 0;
 end;
 *)
 if Message.result <> 0 then inherited;
end;

procedure TCodeEditor.WMVScroll(var Message: TWMVScroll);
begin
  inherited;
  case Message.ScrollCode of
    SB_LINEDOWN: IncFirstLine;
    SB_LINEUP: DecFirstLine;
    SB_PAGEDOWN: IncFirstLine(VisibleLines);
    SB_PAGEUP: DecFirstLine(VisibleLines);
    SB_THUMBPOSITION, SB_THUMBTRACK: SetFirstLine(Message.Pos);
  end;
  Refresh;
end;

procedure TCodeEditor.WMHScroll(var Message: TWMHScroll);
var
  NewFirstChar: Integer;
begin
  inherited;
  NewFirstChar := FFirstChar;
  case Message.ScrollCode of
    SB_LINERIGHT: Inc(NewFirstChar);
    SB_LINELEFT: Dec(NewFirstChar);
    SB_PAGERIGHT: Inc(NewFirstChar);
    SB_PAGELEFT: Dec(NewFirstChar);
    SB_THUMBPOSITION, SB_THUMBTRACK: NewFirstChar := Message.Pos;
  end;
  SetFirstChar(NewFirstChar);
  Refresh;
end;

procedure TCodeEditor.Resize;
begin
  inherited;
  UpdateVertScrollbar;
  FLastChar :=  FFirstChar + ClientWidth div canvas.TextWidth(' ');
  FLastLine:= ZeileBeiPixel(ClientHeight);

  // SetScrollRange(Handle, SB_VERT, 0, FStrings.count - 1, true);
  //SetScrollRange(Handle, SB_HORZ, 0, MaxZeichenAnzahlproZeile-1 , true);
end;


procedure TCodeEditor.WMMoving(var Message:TMessage);
begin
 inherited;
  //TODO: Hinweise bewegen
end;

//Link: Messages
procedure TCodeEditor.WMCopy(var Message:TWMCopy);
begin
  Self.CopytoClipboard;
  Self.Refresh;
  Self.UpdateVertScrollbar;
  inherited;
end;
procedure TCodeEditor.WMPaste(var Message:TWMPaste);
begin
  FUpdatingStrings := true;
  Self.PasteFromClipboard;
  FUpdatingStrings := false;
  Self.Refresh;
  Self.UpdateVertScrollbar;
  inherited;
end;
procedure TCodeEditor.WMCUT (var Message:TWMCut);
begin
  FUpdatingStrings := true;
  Self.CopytoClipboard;
  Self.DeleteSelection;
  FUpdatingStrings := false;
  Self.UpdateVertScrollbar;
  Self.Refresh;
  inherited;
end;
procedure TCodeEditor.WMClear (var Message:TWMClear);
begin
  FUpdatingStrings := true;
  Self.DeleteSelection;
  FUpdatingStrings := false;
  Self.Refresh;
  Self.UpdateVertScrollbar;
  inherited;
end;

class function TCodeEditor.LTest(const S:String;Anfang:Integer=0):String;
var
  I : Integer;
begin   
  result := S;
  I := length(result);
  while (I > Anfang) and (result[I] = ' ') do Dec(I);
  SetLength(result,I);
end;


procedure TCodeEditor.KeyPress(var Key: Char);
 function TapAnzahl:Integer;
 var X,Y:Integer;
 begin
   result := Config.LeerzeichenProTap;
   if not Config.SchlaueTabs then exit;

   Y := FCurY-1;
   if Y = -1 then exit;
   while (length(Quelltext.Zeilen[Y]) = 0) do
   begin
     Dec(Y);
     if Y = -1 then exit;
   end;

   for X := 0 to high(Quelltext.Zeilen[Y]) do
    if ( Quelltext.Zeilen[Y][X].Anfang > FCurX+1) then
    begin
      result := Quelltext.Zeilen[Y][X].Anfang-FCurX-1;
      exit;
    end;
 end;
 function AnzahlderLeerzeichen :Integer;
 var Y:Integer;
 begin
   if FCurY > high(Quelltext.Zeilen) then
   begin
     result := 0;
     exit;
   end;
   Y := FCurY;
   while (length(Quelltext.Zeilen[Y]) = 0) do
   begin
     Y := Y -1;
     if Y = -1 then
     begin
       result := 0;
       exit;
     end;
   end;
   result := Quelltext.Zeilen[Y,0].Anfang-1;
 end;
const
  KEY_BACK=#8;
  KEY_RETURN=#13;
  KEY_PASTE=#22;
  KEY_COPY=#3;
  KEY_SPACE=#32;
  KEY_TAP=#9;
var
  S,StringToInsert:String;
  X,Y:Integer;
  I:Integer;
  LA:Integer;
begin{Link: KeyPress}
 inherited;
 if Key = '(' then
 begin
   if Assigned(FOnFunctionHint) then
   begin
     UndoList.New(TAktion_ZeichenEntfernen.Create(FCurX+1,FCurY));
     FOnFunctionHint('Test',FHintForm.Hint);
     FHintCurPos.X:= FCurX;
     FHintCurPos.Y:= FCurY+1;
     X := FHintCurPos.X ;
     Y := FHintCurPos.Y;
     CurtoPixel(X,Y);
     FHintForm.ShowHint(0,ClientOrigin.x + X,ClientOrigin.y + Y);
   end;
 end;
 case Key of
 (*
 KEY_PASTE:
  begin
   PasteFromClipboard;
  end;
 KEY_COPY:
 begin
  CopyToClipboard;
  exit;
 end;   *)


 KEY_BACK:
  begin
   FUpdatingStrings := true;
   if not((FSelX = FCurX) and (FSelY = FCurY)) then
   begin
    DeleteSelection;
    CheckAnsicht(FCurX,FCurY);
   end
   else
   begin //Nichts Markiert
    if FCurX > 0 then
    begin
     if (FStrings.Count <> 0) and (FCurX <= length(FStrings[FCurY])) then
     begin
      S:= FStrings[FCurY];
      UndoList.New(TAktion_ZeichenEingabe.Create(S[FCurX],FCurX,FCurY));
      Delete(S, FCurX, 1);
      FStrings[FCurY] := LTest(S);
      WortErmittlung(FCurY);
      Quelltext.WortErkennung(FCurY,FCurY);
     end;
     Dec(FCurX);
    end
    else {FCurX = 0}
    begin
     if FCurY > 0 then//Es gibt eine Zeile zu lschen.
     begin//Zeile lschen
      FCurX := length(FStrings[FCurY-1]);
      UndoList.New(TAktion_Return.Create(FCurX,FCurY-1));
      FStrings[FCurY-1] := FStrings[FCurY-1]+FStrings[FCurY];
      FStrings.Delete(FCurY);
      Dec(FCurY);
      //nderungen auch am Quelltext Objekt durchfhren
      WortErmittlung(FCurY);
      Quelltext.DeleteLines(FCurY+1,1);
      Quelltext.WortErkennung(FCurY,FCurY);
     end;
    end;
    CheckAnsicht(FCurX,FCurY);
    FSelX := FCurX;
    FSelY := FCurY;
   end;
   Refresh;//Egal was gemacht wurde-> Neu zeichnen
  end;
 KEY_RETURN:
  begin
   FUpdatingStrings := true;
   while FStrings.Count <= FCurY do FStrings.Add('');
   DeleteSelection;
   LA :=  AnzahlderLeerzeichen;
   if FCurX >= length(FStrings[FCurY]) then
   begin
    FStrings.Insert(FCurY+1,'');
    Quelltext.InsertLines(FCurY+1,1);
    UndoList.New(TAktion_DeleteReturn.Create(FCurX,FCurY,LA));
   end
   else
   begin
    UndoList.New(TAktion_DeleteReturn.Create(FCurX,FCurY,LA));
    S := '';
    //Leerzeichen vorbereiten
    SetLength(S,LA);
    for I := 1 to LA do S[I] := ' ';
    //Unterer Teil
    FStrings.Insert(FCurY+1,LTest(S + copy(FStrings[FCurY],FCurX+1,-FCurX+length(FStrings[FCurY])+1)) );
    //Oberer Teil
    S:= FStrings[FCurY];
    Delete(S,FCurX+1,-FCurX+length(FStrings[FCurY])+1);
    FStrings[FCurY]:= S;
    //nderungen auch am Quelltext objekt vollziehen.
    Quelltext.InsertLines(FCurY,1);//Eine Weitere Zeile wird bentigt.
    WortErmittlung(FCurY);
    WortErmittlung(FCurY+1);
    Quelltext.WortErkennung(FCurY,FCurY+1);
   end;
   FCurX := LA;
   FCurY:= FCurY +1;
   FSelX := FCurX;
   FSelY := FCurY;
   Refresh;
  end;

 KEY_TAP:
  begin
   FUpdatingStrings := true;
   DeleteSelection;
   LA := TapAnzahl;
   if FCurX < length(FStrings[FCurY]) then
   begin
    SetLength(StringToInsert,LA);
    for I := 1 to LA do
    begin
     StringToInsert[I] := ' ';
     UndoList.New(TAktion_ZeichenEntfernen.Create(FCurX+I,FCurY));
    end;
    S:= FStrings[FCurY];
    Insert(StringToInsert,S,FCurX+1);
    FStrings[FCurY]:= S;
    WortErmittlung(FCurY);
    Quelltext.WortErkennung(FCurY,FCurY);
   end;
   Inc(FCurX,LA);
   CheckAnsicht(FCurX,FCurY);//Nur zum Ansicht verschieben
   FSelX := FCurX;
   Refresh;
  end;
 KEY_SPACE:
  begin
   FUpdatingStrings := true;
   DeleteSelection;

   if FCurX < length(FStrings[FCurY]) then
   begin
    UndoList.New(TAktion_ZeichenEntfernen.Create(FCurX+1,FCurY));
    S:= FStrings[FCurY];
    Insert(Key,S,FCurX+1);
    FStrings[FCurY]:= S;
    WortErmittlung(FCurY);
    Quelltext.WortErkennung(FCurY,FCurY);
   end;
   Inc(FCurX);
   CheckAnsicht(FCurX,FCurY);
   FSelX := FCurX;
   Refresh;
  end;

 #33..#255:
  begin
    UndoList.New(TAktion_ZeichenEntfernen.Create(FCurX+1,FCurY));
    FUpdatingStrings := true;
    while FStrings.count <= FCurY do FStrings.Add('');
    while FCurX > length(FStrings[FCurY]) do FStrings[FCurY] := FStrings[FCurY]+' ';
    DeleteSelection;
    S:= FStrings[FCurY];
    Insert(Key,S,FCurX+1);
    FStrings[FCurY]:= S;
    Inc(FCurX);
    CheckAnsicht(FCurX,FCurY);
    WortErmittlung(FCurY);
    Quelltext.WortErkennung(FCurY,FCurY);
    FSelX := FCurX;
    Refresh;
  end;
 end;
 (*
 FSelX := FCurX;
 FSelY := FCurY;
 *)
 ShowNewCurPos;//Neue Cursor Position anzeigen;

 UpdateVertScrollbar;
 FUpdatingStrings := false;
end;

function TCodeEditor.DoMouseWheel(Shift: TShiftState; WheelDelta: Integer; MousePos: TPoint): Boolean;
const
  Scrollgeschwindigkeit=120;
begin
 //TODO Scrollen verbessern
 DecFirstLine( Wheeldelta div (Zeilenabstand*3));

 result := inherited DoMouseWheel(Shift,Wheeldelta,MousePos);
 Refresh;
end;

procedure TCodeEditor.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  ZeileBeiY:Integer;
begin//Link: MouseDown
  inherited;
  SetFocus;
  if Button = mbLeft then
  begin
    if X < FAbstandzumRand then
    begin
      if  (X > Zeilenabstand div 2 - ButtonSize div 2)
      and (X < Zeilenabstand-(Zeilenabstand div 2 - ButtonSize div 2))
      and (Y mod Zeilenabstand > Zeilenabstand div 2-(ButtonSize div 2))
      and (Y mod Zeilenabstand < Zeilenabstand-(Zeilenabstand div 2-(ButtonSize div 2))) then
      //Und dieser auch exsistiert
      begin
        ZeileBeiY := ZeileBeiPixel(Y);
        if (ZeileBeiY < FStrings.Count) and Quelltext.CanMinimize(ZeileBeiY)  then
        begin
          Quelltext.Expand(not Quelltext.IsMinimized(ZeileBeiY),ZeileBeiY);
          SetCursorPos(0,ZeileBeiY);
          Refresh;
        end;
      end
    end
    else
    begin
      if FDoubleClick then exit;
      //Klick auerhalb der Komponente
      FMouseDown:= true;

      FBlinkZeit := 0;
      FTextCursorVisible := true;
      PixeltoChar(X,Y);
      CheckCurPos(X,Y);
      //CheckAnsicht(X,Y); Unntig da es schon bei MouseUp geschieht
      FVerschiebungstattMarkierung := PosRelativetoSelection(X,Y)= 0;
      if FVerschiebungstattMarkierung then
      begin///Verschiebung beginnen
        FVerschiebungsOrt.x :=X;
        FVerschiebungsOrt.Y :=Y;
      end
      else
      begin//Cusor versetzen ; gegebenenfalls Markieren
        FCurY:= Y;
        FCurX:= X;
        if not (ssShift in Shift) then
        begin //Keine Markierung
          FSelX:=FCurX;
          FSelY:=FCurY;
        end;
        ShowNewCurPos;//Neue Cursor Position anzeigen;
      end;
    end
  end;
end;


procedure TCodeEditor.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
 SelStrings:TStrings;
 oldCur:TPoint;
 oldSel:TPoint;
begin
 inherited;
 if Button = mbLeft then
 begin
  if FDoubleClick then
  begin
   FDoubleClick := false;
   FMouseDown := false;
   exit;
  end;
  (*
  //Auerhalb?
  if (X < 0)  or (Y < 0)  or (X > Clientwidth)  or (Y > Clientheight) then
  begin
   FVerschiebungstattMarkierung := false;
   parent.SetFocus;
   exit;
  end;  *)
  PixeltoChar(X,Y);
  CheckCurPos(X,Y);
  CheckAnsicht(X,Y);
  if FMouseDown then
  begin
   FMouseDown:=false;
   if FVerschiebungstattMarkierung then
   begin
    FVerschiebungstattMarkierung := false;
    FUpdatingStrings := true;
    SelStrings := TStringList.create;
    try
     case PosRelativetoSelection(X,Y) of
     -1://Nach oben verschieben
      begin
       UndoList.New(TAktionen.Create(2));
       GetSelectionStrings(SelStrings);
       DeleteSelection;
       FCurX := X;
       FCurY := Y;
       FSelX := X;
       FSelY := Y;
       InsertStrings(SelStrings);
      end;
      1: //Nach unten verschieben
      begin
       UndoList.New(TAktionen.Create(2));
       GetSelectionStrings(SelStrings);
       //Makierung merken und zum Ziel springen
       oldCur.x := FCurX;
       oldCur.y := FCurY;
       oldSel.x := FSelX;
       oldSel.y := FSelY;
       FCurX := X;
       FCurY := Y;
       FSelX := X;
       FSelY := Y;
       //Und einfgen
       InsertStrings(SelStrings);
       //Alte Markierung wiederherstellen;
       FCurX :=oldCur.x  ;
       FCurY :=oldCur.y ;
       FSelX :=oldSel.x  ;
       FSelY  :=oldSel.y ;
       //Markiertes Lschen
       DeleteSelection;
       //Neue Position wiedereinnehmen
       FCurX := X;
       FCurY := Y;
       FSelX := X;
       FSelY := Y;
      end;
      0:
      begin
       FCurX := X;
       FCurY := Y;
       FSelX := X;
       FSelY := Y;
      end;
     end;
     Refresh;
     FUpdatingStrings := false;
    finally
     SelStrings.free;
    end;
   end
   else
   begin //Markierung


    //FSelX := FCurX;
    //FSelY := FCurY;
    FCurX := X;
    FCurY := Y;
    Refresh;
   end;
  end
  else//FMouseDown=false
  begin
   FCurX := X;
   FCUrY := Y;
   FSelX := X;
   FSelY := Y;
  end;
  ShowNewCurPos;//Neue Cursor Position anzeigen;
 end;
end;
//Link: MouseMove
procedure TCodeEditor.MouseMove(Shift: TShiftState; X, Y: Integer);
var
 ZeileBeiY:Integer;
begin
 inherited;
 if X < FAbstandzumRand then
 begin
   //Corser setzen
   Cursor := crArrow;
   {$IFNDEF SPEED3}
   //Wenn sich der Zeiger ber einen Button befindet
   if  (X > Zeilenabstand div 2 - ButtonSize div 2)
   and (X < Zeilenabstand-(Zeilenabstand div 2 - ButtonSize div 2))
   and (Y mod Zeilenabstand > Zeilenabstand div 2-(ButtonSize div 2))
   and (Y mod Zeilenabstand < Zeilenabstand-(Zeilenabstand div 2-(ButtonSize div 2))) then
   //Und dieser auch exsistiert
   begin
     ZeileBeiY := ZeileBeiPixel(Y);
     if ZeileBeiY < FStrings.Count then
     if Quelltext.CanMinimize(ZeileBeiY) then  Cursor := crHandPoint
   end
   {$ENDIF}




 end
 else
  Cursor := crIBeam ;
 if not Focused then exit;
 if ssLeft in Shift then
 begin
  CheckPushedCursor(false);
  Refresh;
 end;
end;

procedure TCodeEditor.CheckPushedCursor(const AnsichtVerschieben:Boolean=true);
const SpeedX=10;
var
  X,Y:Integer;
begin
  FBlinkZeit := 0;
  FTextCursorVisible:= true;
  X := Mouse.CursorPos.x - ClientOrigin.X;
  Y := Mouse.CursorPos.Y - ClientOrigin.Y;
  if AnsichtVerschieben then
  begin
   //Ansicht verschieben
   if X < 0            then FirstChar :=FirstChar + X div SpeedX;
   if X > ClientWidth  then FirstChar :=FirstChar + (X-ClientWidth) div SpeedX;

   if Y < 0            then FirstLine :=FirstLine + Y div Zeilenabstand;
   if Y > ClientHeight then FirstLine :=FirstLine + (Y-ClientHeight) div Zeilenabstand;
  end;
  if FDoubleClick then exit;
  //Markierung durchfhren
  PixeltoChar(X,Y);
  CheckCurPos(X,Y);
  if FVerschiebungstattMarkierung then
  begin//Verschiebung
   FVerschiebungsOrt.x :=X;
   FVerschiebungsOrt.Y :=Y;
  end
  else
  begin//Markierung
   FCurY := Y;
   FCurX := X;
  end;
end;

procedure TCodeEditor.Click;
begin
  SetFocus;
end;

procedure TCodeEditor.UpdateQuelltext;
begin
  Quelltext.free;
  Quelltext := QuelltextKlasse.Create(FStrings);
  Refresh;
end;

procedure  TCodeEditor.Paint;
var
  X,Y,I:Integer;
  CharWidth:Integer;
  SelBeginX,SelBeginY:Integer;
  SelEndX,SelEndY:Integer;

  PixelPosY0:Integer;
  PixelPosY1:Integer;
  PosY0:Integer;
  PosY1:Integer;
  LinesToDraw:Integer;
  Zeile:TZeile;
  MarkierterBereich:TRect;
  ZeichneMarkiertes:Boolean;
begin  //Link: Paint
  //Erstmal die Flche mit Wei fllen
  Canvas.Brush.Color := FConfig.Hintergrund;
  Canvas.Brush.style := bsSolid;
  Canvas.Pen.Style := psClear;
  Canvas.Pen.Mode := pmCopy;
  Canvas.Rectangle(0,0,ClientWidth+1,ClientHeight+1);
  Canvas.Font.Name := 'Courier New';
  Canvas.Font.Pitch := fpFixed;
  Canvas.font.Size := (Zeilenabstand*3) div 4 -2;
  CharWidth := Canvas.TextWidth(' ');
  Zeile := nil;

  //Festellen wie die Selektierung vorliegt.
  GetSelection(SelBeginX,SelBeginY,SelEndX,SelEndY);

  //Korrektur:
  {$IFNDEF KeineAuswahlKorrektur}
  if SelEndY < FStrings.count then
  if SelEndX > length(FStrings[SelEndY]) then
  begin
    SelEndX := length(FStrings[SelEndY]);
    if (SelEndY = SelBeginY) and   (SelBeginX > SelEndX) then SelBeginX := SelEndX;
  end;
  {$ENDIF}

  if Quelltext = nil then exit;


  LinesToDraw:= VisibleLines;


  PosY0 := 0;
  PosY1 := 1;
  PixelPosY0 := PosY0 * Zeilenabstand;
  PixelPosY1 := PosY1 * Zeilenabstand;
  Y := FirstLine;
  while (LinesToDraw > 0) and (Y < FStrings.Count) do
  begin
    Zeile := Quelltext.Zeilen[Y];
    //Canvas wieder richtig einstellen
    Canvas.Brush.Style := bsSolid;
    Canvas.Pen.Style := psClear;
    Canvas.Pen.Mode := pmCopy;



    ZeichneMarkiertes := False;
    if (SelBeginY < Y) and (Y < SelEndY  ) then //Alles Markiert
    begin
      MarkierterBereich.Left := FAbstandzumRand;
      MarkierterBereich.Top := PixelPosY0;
      MarkierterBereich.Right := 300*CharWidth+FAbstandzumRand;
      MarkierterBereich.Bottom := PixelPosY1+1;
      ZeichneMarkiertes := True;
    end
    else {Nicht alles Markiert}
    //Markierung nur in einer Zeile
    if (SelBeginY=Y) and (SelEndY=Y) then
    begin
      if SelBeginX <> SelEndX then
      begin
        MarkierterBereich.Left := FAbstandzumRand + (SelBeginX-FFirstChar)*CharWidth;
        MarkierterBereich.Top := PixelPosY0;
        MarkierterBereich.Right := FAbstandzumRand + (SelEndX-FFirstChar)*CharWidth+1;
        MarkierterBereich.Bottom := PixelPosY1+1;
        ZeichneMarkiertes := True;
      end;
    end
    else
    if (SelBeginY = Y) then //Erste markierte Zeile
    begin
      MarkierterBereich.Left := FAbstandzumRand + (SelBeginX-FFirstChar)*CharWidth;
      MarkierterBereich.Top := PixelPosY0;
      MarkierterBereich.Right := ClientWidth+1;//300*CharWidth+FAbstandzumRand;
      MarkierterBereich.Bottom := PixelPosY1+1;
      ZeichneMarkiertes := True;
    end
    else
    if (SelEndY = Y) then  //Letzte markierte Zeile
    begin
      MarkierterBereich.Left := FAbstandzumRand;
      MarkierterBereich.Top := PixelPosY0;
      MarkierterBereich.Right := FAbstandzumRand + (SelEndX-FFirstChar)*CharWidth+1;
      MarkierterBereich.Bottom := PixelPosY1+1;
      ZeichneMarkiertes := True;
    end;

    //Etwas nicht Markiertes ist sichtbar
    if (SelEndY <= Y) or (SelBeginY >= Y) then
    begin
      Canvas.Brush.Color := Config.FHintergrund;
      for X :=High(Zeile) downto 0 do
      begin
        if  Zeile[X].FontIndex > HighColorIndex then raise Exception.Create('Falscher FontIndex')
        else
        begin
          Canvas.Font.Color := Config.ColorNr[Zeile[X].FontIndex];
          Canvas.Font.style := Config.StyleNr[Zeile[X].FontIndex];
        end;
        Canvas.TextOut(FAbstandzumRand + (Zeile[X].Anfang-1-FFirstChar)*CharWidth,PixelPosY0,Zeile[X].text);
      end
    end;
    //Etwas Markiertes ist sichtbar
    if ZeichneMarkiertes then
    begin
      Canvas.Brush.Color := Config.Hintergrund_Markiert;
      Canvas.Brush.Style := bssolid;
      Canvas.Rectangle(MarkierterBereich);
      Canvas.Brush.Style := bsClear;
      Canvas.Font.Color := Config.MarkierterText;
      for X :=High(Zeile) downto 0 do
      begin
        if  Zeile[X].FontIndex > HighColorIndex then raise Exception.Create('Falscher FontIndex')
        else
          Canvas.Font.style := Config.StyleNr[Zeile[X].FontIndex];
        Canvas.TextRect(MarkierterBereich,(FAbstandzumRand + (Zeile[X].Anfang-1-FFirstChar)*CharWidth),PixelPosY0,Zeile[X].text);
      end
    end;

    //Hinweis zeichnenn: "..end"
    if  Quelltext.CanMinimize(Y) and Quelltext.IsMinimized(Y) then
    begin
      I := FAbstandzumRand + (Zeile[high(Zeile)].Anfang+length(Zeile[high(Zeile)].Text)-1-FFirstChar)*CharWidth;
      Canvas.Pen.Style := psClear;

      if (SelBeginY < Y) and (Y < SelEndY  ) then //Alles Markiert
      begin
        Canvas.Font.Color :=   Config.MarkierterText;
        Canvas.Font.Style := Config.StyleNr[CI_Ausdruck];
        Canvas.Brush.Color :=     Config.FHintergrund_Markiert;
        Canvas.Pen.color := Config.FHintergrund;
      end
      else
      begin
        Canvas.Font.Color :=   Config.ColorNr[CI_Ausdruck];
        Canvas.Font.Style :=  Config.StyleNr[CI_Ausdruck];
        Canvas.Brush.Color :=     Config.FHintergrund;
        Canvas.Pen.color := Config.Hintergrund_Markiert;
      end;
      Canvas.Rectangle(I+2,PixelPosY0,I+Canvas.TextWidth('..end')+6,PixelPosY1);
      Canvas.TextOut(I+4,PixelPosY0,'..end');
      Canvas.Brush.Style := bsclear;
      Canvas.Pen.Style := psSolid;
      Canvas.Pen.Mode := pmCopy;

      Canvas.Pen.width := 1;

      Canvas.Rectangle(I+2,PixelPosY0,I+Canvas.TextWidth('..end')+6,PixelPosY1);

    end;

    //Linken Rand berzeichnen
    Canvas.Pen.Color := Config.Hintergrund;;
    Canvas.Brush.Style := bsSolid;
    Canvas.Brush.Color := Config.Hintergrund;
    Canvas.Rectangle(0,PixelPosY0,FAbstandzumRand,PixelPosY1+1);

    if Quelltext.CanMinimize(Y) then
    begin//Knpfe zeichnen
      Canvas.Pen.Style := psSolid;
      Canvas.Pen.Width := 1;
      Canvas.Pen.Mode := pmNot ;
      Canvas.Brush.Style := bsClear;
      //Rahmen Zeichnen
      Canvas.Rectangle(Zeilenabstand div 2 - ButtonSize div 2,
                       PixelPosY0+(Zeilenabstand div 2-(ButtonSize div 2)),
                       Zeilenabstand-(Zeilenabstand div 2 - ButtonSize div 2),
                       PixelPosY1-(Zeilenabstand div 2-(ButtonSize div 2))
                       );

      //"-" Zeichnen
      Canvas.MoveTo(Zeilenabstand div 2 - ButtonSize div 2+2,PixelPosY0+(Zeilenabstand div 2));
      Canvas.LineTo(Zeilenabstand-(Zeilenabstand div 2 - ButtonSize div 2)-2,PixelPosY0+(Zeilenabstand div 2));

      if Quelltext.IsMinimized(Y) then
      begin //Plusstrich Zeichnen
        Canvas.MoveTo(Zeilenabstand div 2,PixelPosY0+(Zeilenabstand div 2-(ButtonSize div 2))+2);
        Canvas.LineTo(Zeilenabstand div 2,PixelPosY1-(Zeilenabstand div 2-(ButtonSize div 2))-2);

      end;



      Canvas.Brush.Style := bsSolid;
      Canvas.Pen.Style := psClear;
      Canvas.Pen.Mode := pmCopy
    end;



    //Cursor Zeichnen
    begin
      if FVerschiebungstattMarkierung then
      begin
        if (Y = FVerschiebungsOrt.Y) and Focused  then
        begin
          Canvas.Pen.Style := psSolid;
          Canvas.Pen.Width := 2;
          Canvas.Pen.Mode := pmNot ;
          Canvas.MoveTo((FVerschiebungsOrt.X-FFirstChar)*CharWidth+FAbstandzumRand,PixelPosY0);
          Canvas.LineTo((FVerschiebungsOrt.X-FFirstChar)*CharWidth+FAbstandzumRand,PixelPosY1);
        end
      end
      else//Normalerweise:
      if FTextCursorVisible and (Y = FCurY) and Focused then
      begin
        Canvas.Pen.Style := psSolid;
        Canvas.Pen.Width := 2;
        Canvas.Pen.Mode := pmNot ;
        Canvas.MoveTo((FCurX-FFirstChar)*CharWidth+FAbstandzumRand,PixelPosY0+1);
        Canvas.LineTo((FCurX-FFirstChar)*CharWidth+FAbstandzumRand,PixelPosY1-1);
      end;

    //Ganz am Schluss:
    Dec(LinesToDraw);

    //Nchste zu zeichnende Zeile ermitteln
    Y := Quelltext.GetNextLine(Y);

    // Zeichen Position nach unten verschieben
    Inc(PosY0);
    Inc(PosY1);
    PixelPosY0 := PosY0 * Zeilenabstand;
    PixelPosY1 := PosY1 * Zeilenabstand;
    end;
  end;

  Canvas.Pen.Mode := pmNot;
  Canvas.Pen.Style := psSolid;
  Canvas.Pen.Width := 1;
  Canvas.MoveTo(FAbstandzumRand-2,0);
  Canvas.LineTo(FAbstandzumRand-2,ClientHeight);


  if (FStrings.Count = 0) and FTextCursorVisible and Focused then
  begin
    Canvas.Pen.Style := psSolid;
    Canvas.Pen.Width := 2;
    Canvas.Pen.Mode := pmNot ;
    Canvas.MoveTo((FCurX-FFirstChar)*CharWidth+FAbstandzumRand,(FCurY-FFirstLine)*Zeilenabstand+1);
    Canvas.LineTo((FCurX-FFirstChar)*CharWidth+FAbstandzumRand,(FCurY-FFirstLine+1)*Zeilenabstand-1);
    Canvas.Pen.Mode := pmCopy
  end;


  (*
  if (FCurY >= FFirstLine) and (FCurY <= FLastLine) and Focused then
  if FVerschiebungstattMarkierung then
  begin
    Canvas.Pen.Style := psSolid;
    Canvas.Pen.Width := 2;
    Canvas.Pen.Mode := pmNot ;
    Canvas.MoveTo((FVerschiebungsOrt.X-FFirstChar)*CharWidth+FAbstandzumRand,(FVerschiebungsOrt.Y-FFirstLine)*Zeilenabstand);
    Canvas.LineTo((FVerschiebungsOrt.X-FFirstChar)*CharWidth+FAbstandzumRand,(FVerschiebungsOrt.Y-FFirstLine+1)*Zeilenabstand);
    Canvas.Pen.Mode := pmCopy
  end
  else//Normalerweise:
  if FTextCursorVisible then
  begin
    //Canvas.TextOut(FCurX*CharWidth+FAbstandzumRand,(FCurY-FFirstLine)*Zeilenabstand,'|');
    Canvas.Pen.Style := psSolid;
    Canvas.Pen.Width := 2;
    Canvas.Pen.Mode := pmNot ;
    Canvas.MoveTo((FCurX-FFirstChar)*CharWidth+FAbstandzumRand,(FCurY-FFirstLine)*Zeilenabstand+1);
    Canvas.LineTo((FCurX-FFirstChar)*CharWidth+FAbstandzumRand,(FCurY-FFirstLine+1)*Zeilenabstand-1);
    Canvas.Pen.Mode := pmCopy
  end;    *)
end;


procedure TCodeEditor.CheckAnsicht(const X,Y:Integer);
//Sorgt dafr das die angebene Stelle zu sehen ist.
var
  Value:Integer;
  BetragvonValue:Integer;
begin
  if X < FFirstChar then FirstChar := X;
  if X > FLastChar-1 then FirstChar := X+FFirstChar-FLastChar+1;

  Value := 0;
  if Y < FFirstLine then Value  := Y-FirstLine;
  if Y > FLastLine-1 then Value :=FirstLine-FirstLine+Y-(FLastLine-1);

  if Value < 0 then BetragvonValue := -Value else BetragvonValue := Value;

  if BetragvonValue <= 1then
   IncFirstLine(Value)
  else
  begin
    if Y < FFirstLine  then FirstLine := Y;
    if Y > FLastLine-2 then FirstLine := FirstLine+Y-(FLastLine-2);
  end;
  


end;

procedure TCodeEditor.CheckCurPos(var X,Y:Integer);
//berprft ob die Position gltig ist, und korriegiert gegebenenfalls.
//Die Ansicht wird allerdings dabei nicht verschoben.
begin

  if FStrings.Count = 0 then
  begin
    if X < 0 then X := 0;
    Y := 0;
    exit;
  end;
  //Y-Bereicht fr X-berprfung berichtigen
  if Y >= FStrings.Count then Y := FStrings.Count-1;
  if Y < 0 then Y := 0;


  //X-Wert berichtigen

  if X < 0 then
  begin
    X := 0;
    Y := Y-1;
  end;


  //Y-Wert berichtigen
  if Y >= FStrings.Count then Y := FStrings.Count-1;
  if Y < 0 then Y := 0;

end;
procedure TCodeEditor.SetStrings (NewStrings:TStrings);
begin
  FStrings.Assign(NewStrings);
  UpdateQuelltext;
  ShowMessage('Update Quelltext');
end;

procedure TCodeEditor.SetFirstChar(Value:Integer);
const MaxTextLength=300;
begin
 if Value <= 0 then FFirstChar := 0
 else
 if Value > MaxTextLength then FFirstChar := MaxTextLength
 else
 FFirstChar := Value;
 if Value <> FFirstChar then SetScrollPos(Handle, SB_HORZ, FFirstChar, True);
 FLastChar :=  FFirstChar + ClientWidth div canvas.TextWidth(' ');
 SetScrollPos(self.Handle, SB_HORZ, FFirstChar, True);
end;

procedure TCodeEditor.SetFirstLine(Value:Integer);
begin
  if Value < 0 then FFirstLine := 0
  else
  if Value > FStrings.count-1 then
  begin
    if FStrings.count = 0 then FFirstLine := 0 else
    FFirstLine := Quelltext.GetLineOwner(FStrings.count-1) ;
  end
  else
  FFirstLine := Quelltext.GetLineOwner(Value);
  FLastLine:= ZeileBeiPixel(ClientHeight);
  SetScrollPos(self.Handle, SB_VERT, FFirstLine, True);
end;

function TCodeEditor.VisibleLines:Integer;
begin
  result := ClientHeight div Zeilenabstand+1;
end;
//Link: Inc/Dec

procedure TCodeEditor.IncCurY(var CurY:Integer;Value:Integer=1);
var
  Max :Integer;
begin
  if Value < 0 then
  begin
    DecCurY(CurY,-Value);
    exit;
  end;
  Max := FStrings.Count-1;
  while (Value > 0) and (CurY < Max)do
  begin
    CurY := Quelltext.GetNextLine(CurY);
    Dec(Value);
  end;
end;

procedure TCodeEditor.DecCurY(var CurY:Integer;Value:Integer=1);
begin
  if Value < 0 then
  begin
    IncCurY(CurY,-Value);
    exit;
  end;
  while (Value > 0) and (CurY > 0) do
  begin
    CurY := Quelltext.GetLineOwner(CurY-1);
    Dec(Value);
  end;
end;


procedure TCodeEditor.IncFirstLine(Value:Integer=1);
var
  NewFirstLine:Integer;
  Max :Integer;
begin
  if Value < 0 then
  begin
    DecFirstLine(-Value);
    exit;
  end;
  NewFirstLine := FirstLine;
  Max := FStrings.Count-1;
  while (Value > 0) and (NewFirstLine < Max)do
  begin
    NewFirstLine := Quelltext.GetNextLine(NewFirstLine);
    Dec(Value);
  end;
  SetFirstLine(NewFirstLine);
end;

procedure TCodeEditor.DecFirstLine(Value:Integer=1);
var
  NewFirstLine:Integer;
begin
  if Value < 0 then
  begin
    IncFirstLine(-Value);
    exit;
  end;
  NewFirstLine := FirstLine;
  while (Value > 0) and (NewFirstLine > 0) do
  begin
    NewFirstLine := Quelltext.GetLineOwner(NewFirstLine-1);
    Dec(Value);
  end;
  SetFirstLine(NewFirstLine);
end;



procedure TCodeEditor.SetCursorPos(X,Y:Integer);
begin
  FCurX := X;
  FCurY := Y;
  FSelX := FCurX;
  FSelY := FCurY;
  ShowNewCurPos;
end;

function TCodeEditor.ZeileBeiPixel(Pixel_Y:Integer):Integer;
var
  YPos:Integer;
begin
  result := FirstLine;
  YPos := 1;
  while Pixel_Y > YPos*Zeilenabstand do
  begin
    Inc(YPos);
    result := Quelltext.GetNextLine(result);
  end;
end;

//Link: Menu Proceduren
procedure TCodeEditor.Menu_SelectAll;
//Nur fr das Menu
begin
  if FStrings.Count = 0 then exit;
  FCurY := 0;
  FCurX := 0;
  FSelY := FStrings.count-1;
  FSelX := length(FStrings[FSelY]);
  ShowNewCurPos;
  Refresh;
end;

procedure TCodeEditor.Menu_GoTo(X,Y:Integer;NeuZeichnen:Boolean=true);
begin
  FirstLine := Y-3;
  FirstChar := X - (ClientWidth-FAbstandzumRand)*1 div Canvas.TextWidth(' ') div 3;
  if NeuZeichnen then Refresh;
end;

procedure TCodeEditor.Menu_ReplaceText(Y,X,Size:Integer;RText:String;NeuZeichnen:Boolean=true);
var
 RStrings:TStringList;
begin
  FUpdatingStrings := true;
  RStrings := TStringList.Create;
  try
    RStrings.add(RText);
    UndoList.New(TAktionen.Create(2));
    FCurX := X;
    FSelX := X + Size;
    FCurY := Y;
    FSelY := Y;
    DeleteSelection;
    InsertStrings(RStrings);
    FCurX := X;
    FSelX := X + length(RText);
  finally
    RStrings.Free;
  end;
  FUpdatingStrings := false;
  if NeuZeichnen then Refresh;
end;

procedure TCodeEditor.Menu_SelectLine(Y,X,Length:Integer;NeuZeichnen:Boolean=true);
begin
  FCurY := Y;
  FSelY := Y;
  FCurX := X + Length;
  FSelX := X;
  if NeuZeichnen then Refresh;
end;

procedure TCodeEditor.Menu_IncreaseTabs; //Nur fr das Menu
var
  LZ:String;
  SelBeginY,SelEndY:Integer;
  Unwichtig:Integer;
  X,Y:Integer;
begin
  if FStrings.Count = 0 then exit;
  FUpdatingStrings := true;
    GetSelection(Unwichtig,SelBeginY,Unwichtig,SelEndY) ;
    SetLength(LZ,Config.FLeerzeichenProTap);
    for X := 1 to length(LZ) do LZ[X] := ' ';
    for Y := SelBeginY to SelEndY do
    begin
     FStrings[Y] := LZ + FStrings[Y] ;
     WortErmittlung(Y);
    end;
    Quelltext.WortErkennung(SelBeginY,SelEndY);
    Refresh;
  FUpdatingStrings := false;
end;

procedure TCodeEditor.Menu_DecreaseTabs;//Nur fr das Menu
var
  SelBeginY,SelEndY:Integer;
  Unwichtig:Integer;
  Y:Integer;
  S:String;
  TabSize:Integer;
begin
  if FStrings.Count = 0 then exit;
  FUpdatingStrings := true;
    GetSelection(Unwichtig,SelBeginY,Unwichtig,SelEndY) ;
    for Y := SelBeginY to SelEndY do
    if Quelltext.Zeilen[Y] <> nil then
    begin
      S := FStrings[Y];
      TabSize := Config.FLeerzeichenProTap;
      if TabSize > Quelltext.Zeilen[Y][0].Anfang-1 then
       TabSize := Quelltext.Zeilen[Y][0].Anfang-1 ;
      Delete(S,1,TabSize);
      FStrings[Y] := S;
      WortErmittlung(Y);
    end;
    Quelltext.WortErkennung(SelBeginY,SelEndY);
    Refresh;
  FUpdatingStrings := false;
end;

procedure TCodeEditor.Load(const Datei:String);
begin
  QuelltextKlasse := Config.GetQuelltextTyp(Datei);
  Lines.LoadFromFile(Datei);
  LoadedFile := Datei;
end;

procedure TCodeEditor.Menu_SetQuelltextTyp(Typ:TQuelltextKlasse);
begin
  QuelltextKlasse := Typ;
  UpdateQuelltext;
end;

procedure TCodeEditor.ShowNewCurPos;
begin
  Config.NewCurPos(self,FCurX,FCurY);//Verschiebungs Ort angeben
end;

procedure TCodeEditor.Notification(AComponent: TComponent; Operation: TOperation);//Link: Config
begin
  inherited;
  if (Operation=opRemove) and (AComponent=FConfig) and not( csDestroying in ComponentState)  then

  begin
    FConfig := TCodeEditorConfig.Create(self);//Neues Config erstellen
    FConfig.Login(self);
  end;
end;
procedure TCodeEditor.SetConfig(const NewConfig:TCodeEditorConfig);
begin
  if  NewConfig <> FConfig then
  begin
    if Assigned(FConfig) then FConfig.Logout(self);//Alte Verknpfung lschen
    FConfig := NewConfig;
    if Assigned(FConfig) then FConfig.Login(self);//Neue Verknpfung erstellen
  end;
end;

{Config---Config---Config---Config--Config---Config---Config---Config---Config}
{-----------------------------------------------------------------------------}
constructor TCodeEditorConfig.Create(AOwner:TComponent);
begin
  inherited;
  FFileName:='Config.txt';
  FHintergrund := clWindow;
  FMarkierterText:= clHighlightText;
  FHintergrund_Markiert:=clHighlight;
  FLeerzeichenProTap:= StdLeerzeichenProTap;
  FSchlaueTabs:= true;
  ColorNr := StdColorNr; //Standart Werte bernehmen
  StyleNr := StdStyleNr;
end;

procedure TCodeEditorConfig.ReadState(Reader: TReader);
begin
  inherited;
  LoadFromFile(extractFileDir(paramstr(0))+'\'+FFileName);
end;

function TCodeEditorConfig.LoadFromFile(FileName:String):Boolean;
var
  Strings:TStringList;
begin
  result := false;
  Strings := TStringList.Create;
  try
    if FileExists(FileName) then
    begin
      Strings.LoadFromFile(FileName);
      Strings.Sorted := true;
      LoadFromStrings(Strings,[]);
      result := true;
    end;
  finally
    Strings.Free;
  end;
end;

procedure TCodeEditorConfig.LoadFromStrings(Strings:TStrings;ErlaubteReaktion:TLoadConifgReaktion=[LCR_NeuZeichnen,LCR_UpdateQuelltext]);
var
  Reaktion:TLoadConifgReaktion;
procedure LoadByte(var ByteVar:Byte;const Name:String;Aktion:TLoadConifgReaktion=[]);
var
  AByte:Byte;
  AInteger:Integer;
  Text:String;
begin
  Text := Strings.Values[Name];
  if Text <> '' then
  begin
    try
      AInteger := strtoint(Text);
    except
      AInteger := -1;
    end;
    if (AInteger < 0) or (AInteger > 255) then exit;
    AByte := Lo(AInteger);
    if AByte <> ByteVar then
    begin
      ByteVar := AByte;
      Reaktion := Reaktion + Aktion;
    end;
  end;
end;
procedure LoadBoolean(var BooleanVar:Boolean;const Name:String;Aktion:TLoadConifgReaktion=[]);
var
  ABoolean:Boolean;
  Text:String;
begin
  Text := Strings.Values[Name];
  if Text <> '' then
  begin
    Text := LowerCase(Text);
    if Text = 'false' then ABoolean := false
    else
     if Text = 'true' then ABoolean := true
     else
      exit;//Keine nderungen
    if ABoolean <> BooleanVar then
    begin
      BooleanVar := ABoolean;
      Reaktion := Reaktion + Aktion;
    end;
  end;

end;
procedure LoadColor(var ColorVar:TColor;const Name:String);
var
  Color:TColor;
  Text:String;
begin
  Text := Strings.Values[Name];
  if Text <> '' then
  begin
    try
      Color := StringToColor(Text);
    except
      Color := ColorVar;//Farbe bei Fehler
    end;
    if Color <> ColorVar then
    begin
      ColorVar := Color;
      Reaktion := Reaktion + [LCR_NeuZeichnen];
    end;
  end;
end;
begin
  Reaktion:=[];

  LoadColor(FHintergrund,'Hintergrund\Farbe');
  LoadColor(FHintergrund_Markiert,'Hintergrund\Markiert\Farbe');
  LoadColor(FMarkierterText,'Text\Markiert\Farbe');
  LoadColor(ColorNr[CI_Normal],'Text\Standart\Farbe');
  LoadColor(ColorNr[CI_Ausdruck],'Text\Ausdruck\Farbe');
  LoadColor(ColorNr[CI_Kommentar],'Text\Kommentar\Farbe');
  LoadColor(ColorNr[CI_String],'Text\String\Farbe');
  LoadColor(ColorNr[CI_Klammer0],'Text\Klammer0\Farbe');
  LoadColor(ColorNr[CI_Klammer1],'Text\Klammer1\Farbe');
  LoadColor(ColorNr[CI_Klammer2],'Text\Klammer2\Farbe');
  LoadColor(ColorNr[CI_Compiler],'Text\Compiler\Farbe');
  LoadColor(ColorNr[CI_GExperts],'Text\GExperts\Farbe');
  LoadColor(ColorNr[CI_XML_OpenTag_Klammer],'Text\XML\OpenTag\Klammer\Farbe');
  LoadColor(ColorNr[CI_XML_CloseTag_Klammer],'Text\XML\CloseTag\Klammer\Farbe');
  LoadColor(ColorNr[CI_XML_Eigenschaft],'Text\XML\Eigenschaft\Farbe');


  LoadBoolean(FSchlaueTabs,'Tabs\SchlaueTabs',[]);
  LoadByte(FLeerzeichenProTap,'Tabs\Size',[]);
  Reaktion := Reaktion * ErlaubteReaktion;//Schnittmenge

  if  (LCR_NeuZeichnen in Reaktion) then DrawCodeEditors;
end;

function TCodeEditorConfig.GetQuelltextTyp(const Datei:String):TQuelltextKlasse;
var
  Typ:String;
begin
  Typ := ExtractFileExt(Datei);
  if Typ = '.xml' then Result := TXMLQuelltext else Result := TDelphiQuelltext;

end;

procedure TCodeEditorConfig.NewCurPos(const Sender:TCodeEditor;const X,Y:Integer);
begin
  if Assigned(FOnCurPosChange) then FOnCurPosChange(Sender,X,Y);
end;

procedure TCodeEditorConfig.DrawCodeEditors;
var
  I:Integer;
begin
  for I := 0 to high(CodeEditorNr) do CodeEditorNr[I].refresh;
end;

function TCodeEditorConfig.IndexOf(const CodeEditor:TCodeEditor):Integer;
var
  I:Integer;
begin
  result := -1;
  for I := 0 to high(CodeEditorNr) do
   if CodeEditorNr[I] = CodeEditor then result := I;
end;


procedure TCodeEditorConfig.Login(const CodeEditor:TCodeEditor);
var
  I:Integer;
begin
  I := IndexOf(CodeEditor);
  if I = -1 then //Editor ist noch nicht eingeloggt.
  begin
    SetLength(CodeEditorNr,Length(CodeEditorNr)+1);
    CodeEditorNr[high(CodeEditorNr)] := CodeEditor;
  end;
end;

procedure TCodeEditorConfig.Logout(const CodeEditor:TCodeEditor);
var
  I:Integer;
begin
  I := IndexOf(CodeEditor);
  if I <> -1 then //Editor ist eingeloggt
  begin
    CodeEditorNr[I] := CodeEditorNr[high(CodeEditorNr)];
    SetLength(CodeEditorNr,Length(CodeEditorNr)-1);
  end;
end;

procedure TCodeEditorConfig.SetHintergrund(const NewColor:TColor);
begin
  FHintergrund := NewColor;
  DrawCodeEditors;
end;

{--Undo---Undo---Undo---Undo---Undo---Undo---Undo---Undo---Undo---Undo---Undo--}
{------------------------------------------------------------------------------}
//Link:Undo
Constructor TAktionsListe.Create(const Owner:TCodeEditor;const NewSize:Integer);
begin
 inherited Create;
 CodeEditor := Owner;
 FSize := 0;//Anzahl der wirklichen Inhalte
 SetLength(Nr,NewSize);
end;

procedure TAktionsListe.Clear;
var
  I:Integer;
begin
 for I := 0 to FSize-1 do
 begin
   Nr[I].free;
   Nr[I] := nil;
 end;
 FSize := 0;
 FCanUndo := false;
end;

Destructor TAktionsListe.Destroy;
begin
 Clear;
 inherited;
end;


procedure TAktionsListe.Undo;
begin
 if Size <= 1 then FCanUndo := false;
 if Size <= 0 then exit;
 CodeEditor.FUpdatingStrings := true;
 OtherList.New(Nr[Size-1].Execute,false);
 CodeEditor.FUpdatingStrings := false;
 CodeEditor.Refresh;
 Nr[Size-1].Free;
 Nr[Size-1] := nil;
 Dec(FSize);
end;


procedure TUndoList.Add(const Aktion:TAktion);
begin
  if  (Size > 0) then
  begin
    {Aufeinander folgende Zeichen zusammenfassen}
    if ((Aktion is TAktion_ZeichenEntfernen) and (Nr[Size-1] is TAktion_ZeichenEntfernen)       and (Nr[Size-1] as TAktion_ZeichenEntfernen).TryAdd(Aktion as TAktion_ZeichenEntfernen))
    or ((Aktion is TAktion_ZeichenEingabe  ) and (Nr[Size-1] is TAktion_ZeichenEingabe) and (Nr[Size-1] as TAktion_ZeichenEingabe).TryAdd(Aktion as TAktion_ZeichenEingabe)) then
    begin
      Aktion.free;
      exit;
    end;
    {Aktionen zusammenfassen}
    if (Nr[Size-1] is TAktionen) then with Nr[Size-1] as TAktionen do
    begin
      if Anzahl < length(Child) then
      begin
        Child[Anzahl] := Aktion;
        Aktion.FOwner := self;
        Inc(Anzahl);
        exit;
      end;
    end;
  end;
  inherited;
end;



procedure TAktionsListe.Add(const Aktion:TAktion);
var
  I:Integer;
begin
 if FSize < length(Nr) then
 begin
  Nr[FSize] := Aktion;
  Inc(FSize);
 end
 else
 begin
  Nr[0].Free;
  for I := 0 to Size-2 do
  begin
   Nr[I] := Nr[I+1] ;
  end;
  Nr[Size-1] := Aktion;
 end;
 Aktion.FOwner := self;
 FCanUndo := true;
end;

procedure TAktionsListe.New(const NeueAktion:TAktion;ClearOtherList:Boolean=true);
begin
 if NeueAktion = nil then exit;
 if CodeEditor.NoUndo then
 begin
   NeueAktion.free;
   exit;
 end;
 if ClearOtherList then OtherList.Clear;
 self.Add(NeueAktion);
end;




Constructor TAktion_Delete.Create(SelBeginX,SelBeginY,SelEndX,SelEndY:Integer);
begin
 inherited Create;
 SelBegin.X := SelBeginX;
 SelBegin.Y := SelBeginY;
 SelEnd.X := SelEndX;
 SelEnd.Y := SelEndY;
end;

function TAktion_Delete.Execute;
var
  StringstoDelete:TStrings;
  NoUndoModus:Boolean;
begin
  with Owner.CodeEditor do
  begin
    NoUndoModus := NoUndo;//push
    NoUndo := true;
    FCurX :=  SelBegin.X;
    FCurY :=  SelBegin.Y;
    FSelX :=  SelEnd.X;
    FSelY :=  SelEnd.Y;
    StringstoDelete := TStringList.Create;
    GetSelectionStrings(StringstoDelete);
    result := TAktion_Insert.Create(StringstoDelete,SelBegin.X,SelBegin.Y);
    DeleteSelection;
    NoUndo := NoUndoModus;
  end;
  Owner.CodeEditor.ShowNewCurPos;//Neue Cursor Position anzeigen;

end;

function TAktion_Insert.Execute;
var
  NoUndoModus:Boolean;
begin
 if FStrings.Count = 0 then
  begin
   result := nil;
   exit;
  end;
    //31.08: FPosX+ entfernt da es falsche Werte liefert >> Vermutung stimmte!!
    //30.09: Oberes trifft nur auf eine mehrzeilige Markierung zu  Darum:
  if FStrings.Count = 1 then
  begin
    result := TAktion_Delete.Create(FPos.X,FPos.Y,FPos.X+length(FStrings[0]),FPos.Y);
  end
  else
  begin
    result := TAktion_Delete.Create(FPos.X,FPos.Y,length(FStrings[FStrings.count-1]),FPos.Y+FStrings.count-1);
  end;

  with Owner.CodeEditor do
  begin
    //Wichtig: Cursor setzen
    SetCursorPos(FPos.x,FPos.y);
    NoUndoModus := NoUndo;//push
    NoUndo := true;
    InsertStrings(self.FStrings);
    NoUndo := NoUndoModus;//pop
  end;
  Owner.CodeEditor.ShowNewCurPos;//Neue Cursor Position anzeigen;
end;



destructor TAktion_Insert.Destroy;
begin
  FStrings.free;
  inherited;
end;

constructor TAktion_Insert.Create(DeleteStrings:TStrings;PosX,PosY:Integer);
begin
 inherited Create;
 FStrings := DeleteStrings;
 FPos.X := PosX;
 FPos.Y := PosY;
end;

constructor TAktion_ZeichenEntfernen.Create(PosX,PosY:Integer;Anzahl:Integer=1);
begin
 inherited Create;
 Position.x := PosX;
 Position.y := PosY;
 Length := Anzahl;
end;

function TAktion_ZeichenEntfernen.Execute;
var
  S:String;
begin
  S := Owner.CodeEditor.Lines[Position.Y];
  result := TAktion_ZeichenEingabe.Create(copy(S,Position.X,Length),Position.X,Position.Y);
  Delete(S,Position.X,Length);
  S := TCodeEditor.LTest(S);
  Owner.CodeEditor.Lines[Position.Y] := S;
  Owner.CodeEditor.FCurX := system.length(S);
  Owner.CodeEditor.FSelX := system.length(S);
  Owner.CodeEditor.WortErmittlung(Position.Y);
  Owner.CodeEditor.Quelltext.WortErkennung(Position.Y,Position.Y);
  Owner.CodeEditor.ShowNewCurPos;
end;

function TAktion_ZeichenEntfernen.TryAdd(Zeichen:TAktion_ZeichenEntfernen):Boolean;
begin
  if (Zeichen.Position.Y = Position.Y) and (Zeichen.Position.X = Position.X + Length) then
  begin
    result := true;
    Inc(Length);
  end
  else
   result := false;
end;


constructor TAktion_ZeichenEingabe.Create(Zeichen:String;PosX,PosY:Integer);
begin
  inherited Create;
  Position.X := PosX;
  Position.Y := PosY;
  Text := Zeichen;
end;

function TAktion_ZeichenEingabe.Execute;
var
  S:String;
begin
  result := TAktion_ZeichenEntfernen.Create(Position.X,Position.Y,Length(Text));
  S := Owner.CodeEditor.Lines[Position.Y];
  while length(S) < Position.X do S := S + ' ';
  Insert(Text,S,Position.X);
  Owner.CodeEditor.SetCursorPos(Position.X +length(Text),Position.Y);
  Owner.CodeEditor.Lines[Position.Y] := S;
  Owner.CodeEditor.WortErmittlung(Position.Y);
  Owner.CodeEditor.Quelltext.WortErkennung(Position.Y,Position.Y);
end;

function TAktion_ZeichenEingabe.TryAdd(Zeichen:TAktion_ZeichenEingabe):Boolean;
begin
  if (Zeichen.Position.Y = Position.Y) and (Zeichen.Position.X = Position.X-1) then
  begin
    result := true;
    Text := Zeichen.Text + Text;
    Dec(Position.X);
  end
  else
  if (Zeichen.Position.Y = Position.Y) and (Zeichen.Position.X = Position.X) then
  begin
    result := true;
    Text := Text+Zeichen.Text;
  end
  else
   result := false;
end;


constructor TAktion_Return.Create(PosX,PosY:Integer);
begin
  inherited Create;
  FPos.x := PosX;
  FPos.y := PosY;
end;

function TAktion_Return.Execute;
var
  S:String ;
begin
  S:= Owner.CodeEditor.Lines[FPos.Y];
  Owner.CodeEditor.Lines[FPos.Y] := copy(S,FPos.X+1,length(S)-FPos.X);
  SetLength(S,FPos.X);
  //Oberen Teil oberhalb des anderen Teiles einfgen.
  Owner.CodeEditor.Lines.insert(FPos.Y,TCodeEditor.LTest(S));
  result := TAktion_DeleteReturn.Create(FPos.X,FPos.Y,0);
  
  //nderungen auch in Eigenschaft "Quelltext" durchfhren.
  Owner.CodeEditor.Quelltext.InsertLines(FPos.Y,1);
  Owner.CodeEditor.WortErmittlung(FPos.Y);
  Owner.CodeEditor.WortErmittlung(FPos.Y+1);
  Owner.CodeEditor.Quelltext.WortErkennung(FPos.Y,FPos.Y+1);
end;

constructor TAktion_DeleteReturn.Create(const XPos,YPos,LeerzeichenAnzahl:Integer);
begin
  inherited Create;
  FPos.X := XPos;
  FPos.Y := YPos;
  FLAnzahl :=  LeerzeichenAnzahl;
end;


function TAktion_DeleteReturn.Execute;
var
 S1,S2:String;
begin
  with Owner.CodeEditor do
  begin
    S1 := lines[FPos.Y];
    S2 := lines[FPos.Y+1];
    while length(S1) < FPos.X do S1 := S1 + ' ';
    S1 := S1 + Copy(S2,FLAnzahl+1,length(S2)-FLAnzahl);
    Lines.Delete(FPos.Y+1);
    Lines[FPos.Y] := LTest(S1);
    SetCursorPos(FPos.X,FPos.Y);
    //nderungen auch in Eigenschaft "Quelltext" durchfhren.
    Quelltext.DeleteLines(FPos.Y+1,1);
    WortErmittlung(FPos.Y);
    Quelltext.WortErkennung(FPos.Y,FPos.Y);
  end;
  result := TAktion_Return.Create(FPos.X,FPos.Y);
end;

function TAktionen.Execute:TAktion;
var
  I:Integer;
  Max:Integer;
  NeueAktionen:TAktionen absolute result;
begin
  NeueAktionen := TAktionen.Create(length(Child));
  Max:= high(Child);
  for I := Max downto 0 do
  begin
    NeueAktionen.Child[I] := Child[I].Execute;
    NeueAktionen.Child[I].FOwner := self.FOwner;
  end;
end;

constructor TAktionen.Create(const Size:Integer);
begin
  inherited Create;
  Anzahl := 0;
  if Size > 0 then
  begin
   SetLength(Child,Size);
  end
end;

destructor TAktionen.Destroy;
var
  I:Integer;
begin
  for I := 0 to high(Child) do Child[I].free;
  inherited;
end;

procedure Register;
begin
  RegisterComponents('Extras', [TCodeEditor,TCodeEditorConfig]);
end;

end.
