(***** BEGIN LICENSE BLOCK *****
Version: MPL 1.1

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is FindDialogUnit.pas

The Initial Developer of the Original Code is
Florian Kberle. (eMail:FloriansKarten@web.de)
Portions created by the Initial Developer are Copyright (C) 2005
the Initial Developer. All Rights Reserved.

Contributor(s):

***** END LICENSE BLOCK *****)
unit FindDialogUnit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls,CodeEditor, Buttons, ExtCtrls;

type
 TSuchAnfang=(SA_TextAnfang,SA_MarkierungsEnde,SA_MarkierungsAnfang);
 TSuchEnde=  (SE_TextEnde,SE_MarkierungsEnde,SE_MarkierungsAnfang);

type
  TFindForm = class(TForm)
    Edit1: TEdit;
    CheckBoxSize: TCheckBox;
    BitBtn1: TBitBtn;
    BitBtn2: TBitBtn;
    RadioGroupBereich: TRadioGroup;
    RadioGroupStart: TRadioGroup;
    procedure Suchen_Ok_Click(Sender: TObject);
    function Suche(Von:TSuchAnfang;Bis:TSuchEnde):Boolean;
    procedure RadioGroupBereichClick(Sender: TObject);
  private
    { Private-Deklarationen }
  public
    { Public-Deklarationen }
  end;

var
  FindForm: TFindForm;
procedure Weitersuchen;
implementation
uses CodeEditorFormUnit;
{$R *.DFM}

function TFindForm.Suche(Von:TSuchAnfang;Bis:TSuchEnde):Boolean;
var
  X,Y:Integer;
  CEditor:TCodeEditor;
  SuchText:String;
  Text:String;
  VonX,VonY:Integer;
  BisX,BisY:Integer;
  Nothing:Integer;
  high_Lines:Integer;
begin
  result := false;
  SuchText := Edit1.Text;
  if not CheckBoxSize.Checked then SuchText := LowerCase(SuchText);
  CEditor := CodeEditorFormUnit.FloriansCodeEditorForm.GetActiveCodeEditor;
  high_Lines := CEditor.Lines.Count-1;
  if high_Lines = -1 then Exit;

  Case Von of
   SA_TextAnfang:
    begin
      VonX := 0;
      VonY := 0;
    end;
   SA_MarkierungsAnfang:
    CEditor.GetSelection(VonX,VonY,Nothing,Nothing);
   SA_MarkierungsEnde:
    CEditor.GetSelection(Nothing,Nothing,VonX,VonY);
  end;

  Case Bis of
   SE_TextEnde:
    begin
     BisX := 1000000;
     BisY :=  CEditor.Lines.Count-1;
    end;
   SE_MarkierungsAnfang:
    CEditor.GetSelection(BisX,BisY,Nothing,Nothing);
   SE_MarkierungsEnde:
    CEditor.GetSelection(Nothing,Nothing,BisX,BisY);
  end;

  //Im hinteren Teil der ersten Zeile suchen
  Y := VonY;
  Text := CEditor.Lines[Y] ;
  Text := Copy(Text,VonX+1,length(Text));//Vordern Teil wegschneiden
  if not CheckBoxSize.Checked then Text  := LowerCase(Text);
  X := Pos(SuchText,Text);
  if X <> 0 then
  begin
    X := X+VonX;//Den Weggeschnittenen Teil mitzhlen
    CEditor.Menu_GoTo(X-1,Y,false);
    CEditor.Menu_SelectLine(Y,X-1,length(SuchText) );
    result := true;
    exit;
  end;
  //Den Rest durchsuchen;
  Inc(Y);
  if Y > high_Lines then Y := 0;
  while (Y <> BisY) do
  begin
    Text := CEditor.Lines[Y];
    if not CheckBoxSize.Checked then Text  := LowerCase(Text);
    X := Pos(SuchText,Text);
    if (X <> 0) then
    begin
      CEditor.Menu_GoTo(X-1,Y,false);
      CEditor.Menu_SelectLine(Y,X-1,length(SuchText) );
      result := true;
      Exit;
    end;
    Inc(Y);
    if Y > high_Lines then Y := 0;
  end;
  //Letzte Zeile berprfen
  Text := CEditor.Lines[BisY];
  if not CheckBoxSize.Checked then Text  := LowerCase(Text);
  X := Pos(SuchText,Text);
  if (X <> 0) and (X < BisX) then
  begin
    CEditor.Menu_GoTo(X-1,Y,false);
    CEditor.Menu_SelectLine(Y,X-1,length(SuchText) );
    result := true;
    Exit;
  end;
end;

procedure Weitersuchen;
begin
  FindForm.Suche(SA_MarkierungsEnde,SE_MarkierungsEnde);
end;

procedure TFindForm.Suchen_Ok_Click(Sender: TObject);
var
  SuchAnfang:TSuchAnfang;
  SuchEnde:TSuchEnde;
begin
  if RadioGroupBereich.ItemIndex = 0 then  //Gesamter Text
  begin
    if RadioGroupStart.ItemIndex = 0 then //Ab Cursor
    begin
      SuchAnfang := SA_MarkierungsAnfang;
      SuchEnde := SE_MarkierungsAnfang;
    end
    else //TextAnfang
    begin
      SuchAnfang := SA_TextAnfang;
      SuchEnde := SE_TextEnde;
    end;
  end
  else //Markierter Text
  begin
    SuchAnfang := SA_MarkierungsAnfang;
    SuchEnde := SE_MarkierungsEnde;
  end;

  if not Suche(SuchAnfang,SuchEnde) then
  MessageBox(0, 'Suche ergab keine Treffer', 'Hinweis', MB_ICONINFORMATION or MB_OK);
end;

procedure TFindForm.RadioGroupBereichClick(Sender: TObject);
begin
  RadioGroupStart.enabled := (RadioGroupBereich.ItemIndex = 0);
end;

end.
